# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import tarfile
from unittest import mock

from testtools.matchers import Contains, DirExists, Equals, FileExists, Not

import snapcraft
from snapcraft import file_utils
from snapcraft.internal import sources
from snapcraft.plugins.v1 import dotnet
from tests import unit

from . import PluginsV1BaseTestCase


def _setup_dirs(plugin):
    os.makedirs(plugin.builddir)
    open(os.path.join(plugin.builddir, "test-app.xxproj"), "w").close()
    os.makedirs(plugin.installdir)
    executable_path = os.path.join(plugin.installdir, "test-app")
    open(executable_path, "w").close()


class DotNetPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = dotnet.DotNetPlugin.schema()
        self.assertThat(schema, Contains("required"))

    def test_get_pull_properties(self):
        expected_pull_properties = ["dotnet-runtime-version", "dotnet-version"]
        self.assertThat(
            dotnet.DotNetPlugin.get_pull_properties(), Equals(expected_pull_properties)
        )

    def test_get_build_properties(self):
        expected_build_properties = []
        self.assertThat(
            dotnet.DotNetPlugin.get_build_properties(),
            Equals(expected_build_properties),
        )


class DotNetProjectBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            build_attributes = []
            dotnet_runtime_version = dotnet._RUNTIME_DEFAULT
            dotnet_version = dotnet._VERSION_DEFAULT

        self.options = Options()

        # Only amd64 is supported for now.
        patcher = mock.patch(
            "snapcraft.ProjectOptions.deb_arch",
            new_callable=mock.PropertyMock,
            return_value="amd64",
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        def fake_urlopen(request):
            return FakeResponse(request.full_url, checksum)

        class FakeResponse:
            def __init__(self, url: str, checksum: str) -> None:
                self._url = url
                self._checksum = checksum

            def read(self):
                return json.dumps(
                    {
                        "releases": [
                            {
                                "release-version": "2.0.9",
                                "sdk": {
                                    "runtime-version": "2.0.9",
                                    "files": [
                                        {
                                            "name": "dotnet-sdk-linux-x64.tar.gz",
                                            "rid": "linux-x64",
                                            "url": "https://download.microsoft.com/download/f/c/1/fc16c864-b374-4668-83a2-f9f880928b2d/dotnet-sdk-2.1.202-linux-x64.tar.gz",
                                            "hash": "c1b07ce8849619ca505aafd2983bcdd7141536ccae243d4249b0c9665daf107e03a696ad5f1d95560142cd841a0888bbf5f1a8ff77d3bdc3696b5873481f0998",
                                        }
                                    ],
                                },
                            }
                        ]
                    }
                ).encode("utf-8")

        with tarfile.open("test-sdk.tar", "w") as test_sdk_tar:
            open("test-sdk", "w").close()
            test_sdk_tar.add("test-sdk")
        checksum = file_utils.calculate_hash("test-sdk.tar", algorithm="sha512")

        patcher = mock.patch("urllib.request.urlopen")
        urlopen_mock = patcher.start()
        urlopen_mock.side_effect = fake_urlopen
        self.addCleanup(patcher.stop)

        original_check_call = snapcraft.internal.common.run
        patcher = mock.patch("snapcraft.internal.common.run")
        self.mock_check_call = patcher.start()
        self.addCleanup(patcher.stop)

        def side_effect(cmd, *args, **kwargs):
            if cmd[0].endswith("dotnet"):
                pass
            else:
                original_check_call(cmd, *args, **kwargs)

        self.mock_check_call.side_effect = side_effect


class TestDotNetErrors:

    scenarios = (
        (
            "DotNetBadArchitectureError",
            {
                "exception_class": dotnet.DotNetBadArchitectureError,
                "kwargs": {"architecture": "wrong-arch", "supported": ["arch"]},
                "expected_message": (
                    "Failed to prepare the .NET SDK: "
                    "The architecture 'wrong-arch' is not supported. "
                    "Supported architectures are: 'arch'."
                ),
            },
        ),
        (
            "DotNetBadReleaseDataError",
            {
                "exception_class": dotnet.DotNetBadReleaseDataError,
                "kwargs": {"version": "test"},
                "expected_message": (
                    "Failed to prepare the .NET SDK: "
                    "An error occurred while fetching the version details "
                    "for 'test'. Check that the version is correct."
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


class DotNetProjectTest(DotNetProjectBaseTest):
    def test_sdk_in_path(self):
        plugin = dotnet.DotNetPlugin("test-part", self.options, self.project)
        self.assertThat(
            plugin.env(plugin.installdir),
            Contains("PATH={}:$PATH".format(plugin._dotnet_sdk_dir)),
        )
        # Be sure that the PATH doesn't leak into the final snap
        self.assertThat(plugin.env(self.project.stage_dir), Equals([]))
        self.assertThat(plugin.env(self.project.prime_dir), Equals([]))

    def test_init_with_non_amd64_architecture(self):
        with mock.patch(
            "snapcraft.ProjectOptions.deb_arch",
            new_callable=mock.PropertyMock,
            return_value="non-amd64",
        ):
            error = self.assertRaises(
                dotnet.DotNetBadArchitectureError,
                dotnet.DotNetPlugin,
                "test-part",
                self.options,
                self.project,
            )
        self.assertThat(error.architecture, Equals("non-amd64"))

    def test_pull_sdk(self):
        plugin = dotnet.DotNetPlugin("test-part", self.options, self.project)
        with mock.patch.object(sources.Tar, "download", return_value="test-sdk.tar"):
            plugin.pull()

        self.assertThat(
            os.path.join("parts", "test-part", "dotnet", "sdk", "test-sdk"),
            FileExists(),
        )

    def test_clean_pull_removes_dotnet_dir(self):
        dotnet_dir = os.path.join("parts", "test-part", "dotnet", "sdk")
        os.makedirs(dotnet_dir)
        plugin = dotnet.DotNetPlugin("test-part", self.options, self.project)
        plugin.clean_pull()
        self.assertThat(dotnet_dir, Not(DirExists()))

    def test_build_changes_executable_permissions(self):
        plugin = dotnet.DotNetPlugin("test-part", self.options, self.project)
        _setup_dirs(plugin)
        plugin.build()

        self.assertThat(
            os.stat(os.path.join(plugin.installdir, "test-app")).st_mode & 0o777,
            Equals(0o755),
        )


class DotNetProjectBuildCommandsTest(DotNetProjectBaseTest):
    def run_test(self, configuration, build_attributes):
        self.options.build_attributes = build_attributes
        plugin = dotnet.DotNetPlugin("test-part", self.options, self.project)
        _setup_dirs(plugin)
        plugin.build()

        part_dir = os.path.join(self.path, "parts", "test-part")
        dotnet_command = os.path.join(part_dir, "dotnet", "sdk", "dotnet")
        self.assertThat(
            self.mock_check_call.mock_calls,
            Equals(
                [
                    mock.call(
                        [dotnet_command, "build", "-c", configuration], cwd=mock.ANY
                    ),
                    mock.call(
                        [
                            dotnet_command,
                            "publish",
                            "-c",
                            configuration,
                            "-o",
                            plugin.installdir,
                            "--self-contained",
                            "-r",
                            "linux-x64",
                        ],
                        cwd=mock.ANY,
                    ),
                ]
            ),
        )

    def test_debug(self):
        self.run_test("Debug", ["debug"])

    def test_release(self):
        self.run_test("Release", ["release"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import doctest
import platform
from unittest import mock

from testscenarios import multiply_scenarios

from snapcraft import project
from snapcraft.internal import repo as snapcraft_repo
from snapcraft.internal.project_loader.grammar_processing import PartGrammarProcessor
from snapcraft.internal.project_loader.grammar_processing import (
    _part_grammar_processor as processor,
)


def load_tests(loader, tests, ignore):
    if project.Project().deb_arch == "amd64":
        tests.addTests(doctest.DocTestSuite(processor))
    return tests


class TestPartGrammarSource:

    source_scenarios = [
        (
            "empty",
            {
                "properties": {"source": ""},
                "expected_amd64": "",
                "expected_i386": "",
                "expected_armhf": "",
            },
        ),
        (
            "plain string",
            {
                "properties": {"source": "foo"},
                "expected_amd64": "foo",
                "expected_i386": "foo",
                "expected_armhf": "foo",
            },
        ),
        (
            "on amd64",
            {
                "properties": {"source": [{"on amd64": "foo"}]},
                "expected_amd64": "foo",
                "expected_i386": "",
                "expected_armhf": "foo",  # 'on' cares about host arch, not target
            },
        ),
        (
            "on i386",
            {
                "properties": {"source": [{"on i386": "foo"}]},
                "expected_amd64": "",
                "expected_i386": "foo",
                "expected_armhf": "",
            },
        ),
        (
            "on armhf",
            {
                "properties": {"source": [{"on armhf": "foo"}]},
                "expected_amd64": "",
                "expected_i386": "",
                "expected_armhf": "",
            },
        ),
        (
            "on amd64 with else",
            {
                "properties": {"source": [{"on amd64": "foo"}, {"else": "bar"}]},
                "expected_amd64": "foo",
                "expected_i386": "bar",
                "expected_armhf": "foo",  # 'on' cares about host arch, not target
            },
        ),
        (
            "on i386 with else",
            {
                "properties": {"source": [{"on i386": "foo"}, {"else": "bar"}]},
                "expected_amd64": "bar",
                "expected_i386": "foo",
                "expected_armhf": "bar",
            },
        ),
        # 'on' only cares about the host arch, not target
        (
            "on armhf with else",
            {
                "properties": {"source": [{"on armhf": "foo"}, {"else": "bar"}]},
                "expected_amd64": "bar",
                "expected_i386": "bar",
                "expected_armhf": "bar",
            },
        ),
        (
            "try",
            {
                "properties": {"source": [{"try": "foo"}]},
                "expected_amd64": "foo",
                "expected_i386": "foo",
                "expected_armhf": "foo",
            },
        ),
        (
            "to amd64",
            {
                "properties": {"source": [{"to amd64": "foo"}]},
                "expected_amd64": "foo",
                "expected_i386": "",
                "expected_armhf": "",
            },
        ),
        (
            "to i386",
            {
                "properties": {"source": [{"to i386": "foo"}]},
                "expected_amd64": "",
                "expected_i386": "foo",
                "expected_armhf": "",
            },
        ),
        (
            "to armhf",
            {
                "properties": {"source": [{"to armhf": "foo"}]},
                "expected_amd64": "",
                "expected_i386": "",
                "expected_armhf": "foo",
            },
        ),
        (
            "to amd64 with else",
            {
                "properties": {"source": [{"to amd64": "foo"}, {"else": "bar"}]},
                "expected_amd64": "foo",
                "expected_i386": "bar",
                "expected_armhf": "bar",
            },
        ),
        (
            "to i386 with else",
            {
                "properties": {"source": [{"to i386": "foo"}, {"else": "bar"}]},
                "expected_amd64": "bar",
                "expected_i386": "foo",
                "expected_armhf": "bar",
            },
        ),
        (
            "to armhf with else",
            {
                "properties": {"source": [{"to armhf": "foo"}, {"else": "bar"}]},
                "expected_amd64": "bar",
                "expected_i386": "bar",
                "expected_armhf": "foo",
            },
        ),
        (
            "on amd64 to armhf",
            {
                "properties": {"source": [{"on amd64 to armhf": "foo"}]},
                "expected_amd64": "",
                "expected_i386": "",
                "expected_armhf": "foo",
            },
        ),
        (
            "on amd64 to armhf with else",
            {
                "properties": {
                    "source": [{"on amd64 to armhf": "foo"}, {"else": "bar"}]
                },
                "expected_amd64": "bar",
                "expected_i386": "bar",
                "expected_armhf": "foo",
            },
        ),
    ]

    arch_scenarios = [
        ("amd64", {"host_arch": "x86_64", "target_arch": "amd64"}),
        ("i386", {"host_arch": "i686", "target_arch": "i386"}),
        ("amd64 to armhf", {"host_arch": "x86_64", "target_arch": "armhf"}),
    ]

    scenarios = multiply_scenarios(source_scenarios, arch_scenarios)

    def test(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        properties,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        repo = mock.Mock()
        plugin = mock.Mock()
        plugin.properties = properties.copy()

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert (
            PartGrammarProcessor(
                plugin=plugin,
                properties=plugin.properties,
                project=project.Project(target_deb_arch=target_arch),
                repo=repo,
            ).get_source()
            == expected_arch[f"expected_{target_arch}"]
        )

        # Verify that the original properties haven't changed
        assert plugin.properties == properties


class TestPartGrammarBuildAndStageSnaps:

    source_scenarios = [
        (
            "empty",
            {
                "snaps": "",
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": set(),
            },
        ),
        (
            "single item",
            {
                "snaps": ["foo"],
                "expected_amd64": {"foo"},
                "expected_i386": {"foo"},
                "expected_armhf": {"foo"},
            },
        ),
        (
            "on amd64",
            {
                "snaps": [{"on amd64": ["foo"]}],
                "expected_amd64": {"foo"},
                "expected_i386": set(),
                "expected_armhf": {"foo"},  # 'on' cares about host, not target
            },
        ),
        (
            "try",
            {
                "snaps": [{"try": ["hello"]}],
                "expected_amd64": {"hello"},
                "expected_i386": {"hello"},
                "expected_armhf": {"hello"},
            },
        ),
        (
            "try optional",
            {
                "snaps": [{"try": ["-invalid-"]}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": set(),
            },
        ),
        (
            "to armhf",
            {
                "snaps": [{"to armhf": ["foo"]}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo"},
            },
        ),
        (
            "on amd64 to armhf",
            {
                "snaps": [{"on amd64 to armhf": "foo"}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo"},
            },
        ),
    ]

    arch_scenarios = [
        ("amd64", {"host_arch": "x86_64", "target_arch": "amd64"}),
        ("i386", {"host_arch": "i686", "target_arch": "i386"}),
        ("amd64 to armhf", {"host_arch": "x86_64", "target_arch": "armhf"}),
    ]

    scenarios = multiply_scenarios(source_scenarios, arch_scenarios)

    def test_snaps(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        snaps,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))
        monkeypatch.setattr(
            snapcraft_repo.snaps.SnapPackage,
            "is_valid_snap",
            lambda x: "invalid" not in x,
        )

        repo = mock.Mock()

        class Plugin:
            build_snaps = snaps
            stage_snaps = snaps

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={
                "build-snaps": {"plugin-preferred"},
                "stage-snaps": "plugin-preferred",
            },
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert processor.get_build_snaps() == expected_arch[f"expected_{target_arch}"]
        assert processor.get_stage_snaps() == expected_arch[f"expected_{target_arch}"]

    def test_snaps_no_plugin_attribute(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        snaps,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))
        monkeypatch.setattr(
            snapcraft_repo.snaps.SnapPackage,
            "is_valid_snap",
            lambda x: "invalid" not in x,
        )

        repo = mock.Mock()

        class Plugin:
            pass

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={"build-snaps": snaps, "stage-snaps": snaps},
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert processor.get_build_snaps() == expected_arch[f"expected_{target_arch}"]
        assert processor.get_stage_snaps() == expected_arch[f"expected_{target_arch}"]


class TestPartGrammarStagePackages:

    source_scenarios = [
        (
            "empty",
            {
                "packages": "",
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": set(),
            },
        ),
        (
            "single item",
            {
                "packages": ["foo"],
                "expected_amd64": {"foo"},
                "expected_i386": {"foo"},
                "expected_armhf": {"foo"},
            },
        ),
        (
            "on amd64",
            {
                "packages": [{"on amd64": ["foo"]}],
                "expected_amd64": {"foo"},
                "expected_i386": set(),
                "expected_armhf": {"foo"},  # 'on' cares about host, not target
            },
        ),
        (
            "try",
            {
                "packages": [{"try": ["foo"]}],
                "expected_amd64": {"foo"},
                "expected_i386": {"foo"},
                "expected_armhf": {"foo"},
            },
        ),
        (
            "to armhf",
            {
                "packages": [{"to armhf": ["foo"]}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo:armhf"},
            },
        ),
        (
            "on amd64 to armhf",
            {
                "packages": [{"on amd64 to armhf": "foo"}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo:armhf"},
            },
        ),
    ]

    arch_scenarios = [
        ("amd64", {"host_arch": "x86_64", "target_arch": "amd64"}),
        ("i386", {"host_arch": "i686", "target_arch": "i386"}),
        ("amd64 to armhf", {"host_arch": "x86_64", "target_arch": "armhf"}),
    ]

    scenarios = multiply_scenarios(source_scenarios, arch_scenarios)

    def test_packages(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        packages,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        repo = mock.Mock()

        class Plugin:
            build_packages = packages
            stage_packages = packages

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={"stage-packages": "plugin-preferred"},
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert (
            processor.get_stage_packages() == expected_arch[f"expected_{target_arch}"]
        )

    def test_packages_plugin_no_attr(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        packages,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        repo = mock.Mock()

        class Plugin:
            pass

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={"stage-packages": packages},
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert (
            processor.get_stage_packages() == expected_arch[f"expected_{target_arch}"]
        )


class TestPartGrammarBuildPackages:

    source_scenarios = [
        (
            "empty",
            {
                "packages": "",
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": set(),
            },
        ),
        (
            "single item",
            {
                "packages": ["foo"],
                "expected_amd64": {"foo"},
                "expected_i386": {"foo"},
                "expected_armhf": {"foo"},
            },
        ),
        (
            "on amd64",
            {
                "packages": [{"on amd64": ["foo"]}],
                "expected_amd64": {"foo"},
                "expected_i386": set(),
                "expected_armhf": {"foo"},  # 'on' cares about host, not target
            },
        ),
        (
            "try",
            {
                "packages": [{"try": ["foo"]}],
                "expected_amd64": {"foo"},
                "expected_i386": {"foo"},
                "expected_armhf": {"foo"},
            },
        ),
        (
            "to armhf",
            {
                "packages": [{"to armhf": ["foo"]}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo"},
            },
        ),
        (
            "on amd64 to armhf",
            {
                "packages": [{"on amd64 to armhf": "foo"}],
                "expected_amd64": set(),
                "expected_i386": set(),
                "expected_armhf": {"foo"},
            },
        ),
    ]

    arch_scenarios = [
        ("amd64", {"host_arch": "x86_64", "target_arch": "amd64"}),
        ("i386", {"host_arch": "i686", "target_arch": "i386"}),
        ("amd64 to armhf", {"host_arch": "x86_64", "target_arch": "armhf"}),
    ]

    scenarios = multiply_scenarios(source_scenarios, arch_scenarios)

    def test_packages(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        packages,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        repo = mock.Mock()

        class Plugin:
            build_packages = packages
            stage_packages = packages

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={"build-packages": {"plugin-preferred"}},
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert (
            processor.get_build_packages() == expected_arch[f"expected_{target_arch}"]
        )

    def test_packages_plugin_no_attr(
        self,
        monkeypatch,
        host_arch,
        target_arch,
        packages,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        repo = mock.Mock()

        class Plugin:
            pass

        plugin = Plugin()
        processor = PartGrammarProcessor(
            plugin=plugin,
            properties={"build-packages": packages},
            project=project.Project(target_deb_arch=target_arch),
            repo=repo,
        )

        expected_arch = dict(
            expected_amd64=expected_amd64,
            expected_i386=expected_i386,
            expected_armhf=expected_armhf,
        )
        assert (
            processor.get_build_packages() == expected_arch[f"expected_{target_arch}"]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals, Not

from snapcraft.extractors._metadata import ExtractedMetadata
from tests import unit


class ExtractedMetadataTestCase(unit.TestCase):
    def test_init(self):
        metadata = ExtractedMetadata(summary="summary")

        self.assertThat(metadata.get_summary(), Equals("summary"))
        self.assertThat(metadata.get_description(), Equals(None))

    def test_update_merge(self):
        metadata = ExtractedMetadata(summary="summary")
        metadata2 = ExtractedMetadata(description="description")
        metadata.update(metadata2)

        self.assertThat(metadata.get_summary(), Equals("summary"))
        self.assertThat(metadata.get_description(), Equals("description"))

    def test_update_overwrite(self):
        metadata = ExtractedMetadata(summary="summary", description="description")
        metadata2 = ExtractedMetadata(description="new description")
        metadata.update(metadata2)

        self.assertThat(metadata.get_summary(), Equals("summary"))
        self.assertThat(metadata.get_description(), Equals("new description"))

    def test_overlap(self):
        metadata = ExtractedMetadata(summary="summary", description="description")
        metadata2 = ExtractedMetadata(description="new description")

        self.assertThat(metadata.overlap(metadata2), Equals({"description"}))

    def test_eq(self):
        metadata1 = ExtractedMetadata(summary="summary")
        metadata2 = ExtractedMetadata(summary="summary")
        self.assertThat(metadata1, Equals(metadata2))

    def test_empty_eq(self):
        self.assertThat(ExtractedMetadata(), Equals(ExtractedMetadata()))

    def test_not_eq(self):
        metadata1 = ExtractedMetadata(summary="summary")
        metadata2 = ExtractedMetadata(description="description")
        self.assertThat(metadata1, Not(Equals(metadata2)))

    def test_len(self):
        metadata = ExtractedMetadata(version="version")
        self.assertThat(len(metadata), Equals(1))

        metadata = ExtractedMetadata(summary="summary", version="version")
        self.assertThat(len(metadata), Equals(2))

    def test_to_dict_partial(self):
        metadata = ExtractedMetadata(summary="summary")
        self.assertThat(metadata.to_dict(), Equals({"summary": "summary"}))

    def test_to_dict_complete(self):
        metadata = ExtractedMetadata(summary="summary", description="description")
        self.assertThat(
            metadata.to_dict(),
            Equals({"summary": "summary", "description": "description"}),
        )

    def test_to_dict_is_a_copy(self):
        metadata = ExtractedMetadata(summary="summary")
        metadata_dict = metadata.to_dict()
        metadata_dict["summary"] = "edited summary"

        # Ensure the metadata cannot be edited with its dict
        self.assertThat(metadata.get_summary(), Equals("summary"))


class TestExtractedMetadataGetters:

    scenarios = [
        ("common_id", {"prop": "common_id", "value": "test-value"}),
        ("summary", {"prop": "summary", "value": "test-value"}),
        ("description", {"prop": "description", "value": "test-value"}),
        ("version", {"prop": "version", "value": "test-value"}),
        ("grade", {"prop": "grade", "value": "test-value"}),
        ("icon", {"prop": "icon", "value": "test-value"}),
        ("desktop_file_paths", {"prop": "desktop_file_paths", "value": ["test-value"]}),
    ]

    properties = (
        "common_id",
        "summary",
        "description",
        "version",
        "grade",
        "icon",
        "desktop_file_paths",
    )

    def test_getters(self, tmp_work_path, prop, value):
        metadata = ExtractedMetadata(**{prop: value})
        for p in self.properties:
            gotten = getattr(metadata, "get_{}".format(p))()
            if p == prop:
                assert gotten == value
            else:
                assert gotten is None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import logging
import re
from copy import deepcopy
from typing import Any, Dict, List, Optional

from . import errors

logger = logging.getLogger(__name__)


class PackageRepository(abc.ABC):
    @abc.abstractmethod
    def marshal(self) -> Dict[str, Any]:
        ...

    @classmethod
    def unmarshal(cls, data: Dict[str, str]) -> "PackageRepository":
        if not isinstance(data, dict):
            raise errors.PackageRepositoryValidationError(
                url=str(data),
                brief=f"Invalid package repository object {data!r}.",
                details="Package repository must be a valid dictionary object.",
                resolution="Verify repository configuration and ensure that the correct syntax is used.",
            )

        if "ppa" in data:
            return PackageRepositoryAptPpa.unmarshal(data)

        return PackageRepositoryApt.unmarshal(data)

    @classmethod
    def unmarshal_package_repositories(cls, data: Any) -> List["PackageRepository"]:
        repositories = list()

        if data is not None:
            if not isinstance(data, list):
                raise errors.PackageRepositoryValidationError(
                    url=str(data),
                    brief=f"Invalid package-repositories list object {data!r}.",
                    details="Package repositories must be a list of objects.",
                    resolution="Verify 'package-repositories' configuration and ensure that the correct syntax is used.",
                )

            for repository in data:
                package_repo = cls.unmarshal(repository)
                repositories.append(package_repo)

        return repositories


class PackageRepositoryAptPpa(PackageRepository):
    def __init__(self, *, ppa: str) -> None:
        self.type = "apt"
        self.ppa = ppa

        self.validate()

    def marshal(self) -> Dict[str, Any]:
        data = dict(type="apt")
        data["ppa"] = self.ppa
        return data

    def validate(self) -> None:
        if not self.ppa:
            raise errors.PackageRepositoryValidationError(
                url=self.ppa,
                brief=f"Invalid PPA {self.ppa!r}.",
                details="PPAs must be non-empty strings.",
                resolution="Verify repository configuration and ensure that 'ppa' is correctly specified.",
            )

    @classmethod
    def unmarshal(cls, data: Dict[str, str]) -> "PackageRepositoryAptPpa":
        if not isinstance(data, dict):
            raise errors.PackageRepositoryValidationError(
                url=str(data),
                brief=f"Invalid package repository object {data!r}.",
                details="Package repository must be a valid dictionary object.",
                resolution="Verify repository configuration and ensure that the correct syntax is used.",
            )

        data_copy = deepcopy(data)

        ppa = data_copy.pop("ppa", "")
        repo_type = data_copy.pop("type", None)

        if repo_type != "apt":
            raise errors.PackageRepositoryValidationError(
                url=ppa,
                brief=f"Unsupported type {repo_type!r}.",
                details="The only currently supported type is 'apt'.",
                resolution="Verify repository configuration and ensure that 'type' is correctly specified.",
            )

        if not isinstance(ppa, str):
            raise errors.PackageRepositoryValidationError(
                url=ppa,
                brief=f"Invalid PPA {ppa!r}.",
                details="PPA must be a valid string.",
                resolution="Verify repository configuration and ensure that 'ppa' is correctly specified.",
            )

        if data_copy:
            keys = ", ".join([repr(k) for k in data_copy.keys()])
            raise errors.PackageRepositoryValidationError(
                url=ppa,
                brief=f"Found unsupported package repository properties {keys}.",
                resolution="Verify repository configuration and ensure that it is correct.",
            )

        return cls(ppa=ppa)


class PackageRepositoryApt(PackageRepository):
    def __init__(
        self,
        *,
        architectures: Optional[List[str]] = None,
        components: Optional[List[str]] = None,
        formats: Optional[List[str]] = None,
        key_id: str,
        key_server: Optional[str] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
        suites: Optional[List[str]] = None,
        url: str,
    ) -> None:
        self.type = "apt"
        self.architectures = architectures
        self.components = components
        self.formats = formats
        self.key_id = key_id
        self.key_server = key_server

        if name is None:
            # Default name is URL, stripping non-alphanumeric characters.
            self.name: str = re.sub(r"\W+", "_", url)
        else:
            self.name = name

        self.path = path
        self.suites = suites
        self.url = url

        self.validate()

    def marshal(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {"type": "apt"}

        if self.architectures:
            data["architectures"] = self.architectures

        if self.components:
            data["components"] = self.components

        if self.formats:
            data["formats"] = self.formats

        data["key-id"] = self.key_id

        if self.key_server:
            data["key-server"] = self.key_server

        data["name"] = self.name

        if self.path:
            data["path"] = self.path

        if self.suites:
            data["suites"] = self.suites

        data["url"] = self.url

        return data

    def validate(self) -> None:  # noqa: C901
        if self.formats is not None:
            for repo_format in self.formats:
                if repo_format not in ["deb", "deb-src"]:
                    raise errors.PackageRepositoryValidationError(
                        url=self.url,
                        brief=f"Invalid format {repo_format!r}.",
                        details="Valid formats include: deb and deb-src.",
                        resolution="Verify the repository configuration and ensure that 'formats' is correctly specified.",
                    )

        if not self.key_id or not re.match(r"^[0-9A-F]{40}$", self.key_id):
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"Invalid key identifier {self.key_id!r}.",
                details="Key IDs must be 40 upper-case hex characters.",
                resolution="Verify the repository configuration and ensure that 'key-id' is correctly specified.",
            )

        if not self.url:
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"Invalid URL {self.url!r}.",
                details="URLs must be non-empty strings.",
                resolution="Verify the repository configuration and ensure that 'url' is correctly specified.",
            )

        if self.suites:
            for suite in self.suites:
                if suite.endswith("/"):
                    raise errors.PackageRepositoryValidationError(
                        url=self.url,
                        brief=f"Invalid suite {suite!r}.",
                        details="Suites must not end with a '/'.",
                        resolution="Verify the repository configuration and remove the trailing '/ from suites or use the 'path' property to define a path.",
                    )

        if self.path is not None and self.path == "":
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"Invalid path {self.path!r}.",
                details="Paths must be non-empty strings.",
                resolution="Verify the repository configuration and ensure that 'path' is a non-empty string such as '/'.",
            )

        if self.path and self.components:
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"Components {self.components!r} cannot be combined with path {self.path!r}.",
                details="Path and components are incomptiable options.",
                resolution="Verify the repository configuration and remove 'path' or 'components'.",
            )

        if self.path and self.suites:
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"Suites {self.suites!r} cannot be combined with path {self.path!r}.",
                details="Path and suites are incomptiable options.",
                resolution="Verify the repository configuration and remove 'path' or 'suites'.",
            )

        if self.suites and not self.components:
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"No components specified.",
                details="Components are required when using suites.",
                resolution="Verify the repository configuration and ensure that 'components' is correctly specified.",
            )

        if self.components and not self.suites:
            raise errors.PackageRepositoryValidationError(
                url=self.url,
                brief=f"No suites specified.",
                details="Suites are required when using components.",
                resolution="Verify the repository configuration and ensure that 'suites' is correctly specified.",
            )

    @classmethod  # noqa: C901
    def unmarshal(cls, data: Dict[str, Any]) -> "PackageRepositoryApt":
        if not isinstance(data, dict):
            raise errors.PackageRepositoryValidationError(
                url=str(data),
                brief=f"Invalid package repository object {data!r}.",
                details="Package repository must be a valid dictionary object.",
                resolution="Verify repository configuration and ensure that the correct syntax is used.",
            )

        data_copy = deepcopy(data)

        architectures = data_copy.pop("architectures", None)
        components = data_copy.pop("components", None)
        formats = data_copy.pop("formats", None)
        key_id = data_copy.pop("key-id", None)
        key_server = data_copy.pop("key-server", None)
        name = data_copy.pop("name", None)
        path = data_copy.pop("path", None)
        suites = data_copy.pop("suites", None)
        url = data_copy.pop("url", "")
        repo_type = data_copy.pop("type", None)

        if repo_type != "apt":
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Unsupported type {repo_type!r}.",
                details="The only currently supported type is 'apt'.",
                resolution="Verify repository configuration and ensure that 'type' is correctly specified.",
            )

        if architectures is not None and (
            not isinstance(architectures, list)
            or not all(isinstance(x, str) for x in architectures)
        ):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid architectures {architectures!r}.",
                details="Architectures must be a list of valid architecture strings.",
                resolution="Verify repository configuration and ensure that 'architectures' is correctly specified.",
            )

        if components is not None and (
            not isinstance(components, list)
            or not all(isinstance(x, str) for x in components)
            or not components
        ):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid components {components!r}.",
                details="Components must be a list of strings.",
                resolution="Verify repository configuration and ensure that 'components' is correctly specified.",
            )

        if formats is not None and (
            not isinstance(formats, list)
            or not all(isinstance(x, str) for x in formats)
        ):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid formats {formats!r}.",
                details="Formats must be a list of strings.",
                resolution="Verify repository configuration and ensure that 'formats' is correctly specified.",
            )

        if not isinstance(key_id, str):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid key identifier {key_id!r}.",
                details="Key identifiers must be a valid string.",
                resolution="Verify repository configuration and ensure that 'key-id' is correctly specified.",
            )

        if key_server is not None and not isinstance(key_server, str):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid key server {key_server!r}.",
                details="Key servers must be a valid string.",
                resolution="Verify repository configuration and ensure that 'key-server' is correctly specified.",
            )

        if name is not None and not isinstance(name, str):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid name {name!r}.",
                details="Names must be a valid string.",
                resolution="Verify repository configuration and ensure that 'name' is correctly specified.",
            )

        if path is not None and not isinstance(path, str):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid path {path!r}.",
                details="Paths must be a valid string.",
                resolution="Verify repository configuration and ensure that 'path' is correctly specified.",
            )

        if suites is not None and (
            not isinstance(suites, list)
            or not all(isinstance(x, str) for x in suites)
            or not suites
        ):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid suites {suites!r}.",
                details="Suites must be a list of strings.",
                resolution="Verify repository configuration and ensure that 'suites' is correctly specified.",
            )

        if not isinstance(url, str):
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Invalid URL {url!r}.",
                details="URLs must be a valid string.",
                resolution="Verify repository configuration and ensure that 'url' is correctly specified.",
            )

        if data_copy:
            keys = ", ".join([repr(k) for k in data_copy.keys()])
            raise errors.PackageRepositoryValidationError(
                url=url,
                brief=f"Found unsupported package repository properties {keys}.",
                resolution="Verify repository configuration and ensure it is correct.",
            )

        return cls(
            architectures=architectures,
            components=components,
            formats=formats,
            key_id=key_id,
            key_server=key_server,
            name=name,
            suites=suites,
            url=url,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess

from testtools.matchers import Contains

from tests import integration


class MultiarchTestCase(integration.TestCase):
    def test_bad_multiarch_stage_packages(self):
        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "pull", "bad-arch"
        )

        self.assertThat(
            exception.output,
            Contains(
                "Failed to fetch stage packages: "
                "Error downloading packages for part 'my-part': The "
                "package 'hello:fake-arch' was not found.\n"
                "You may need to add support for this architecture with "
                "'dpkg --add-architecture fake-arch'"
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

import pytest

from snapcraft.internal.project_loader._extensions.flutter_dev import (
    ExtensionImpl as FlutterDevExtension,
)
from snapcraft.internal.project_loader._extensions.flutter_master import (
    ExtensionImpl as FlutterMasterExtension,
)


@pytest.fixture(params=[FlutterMasterExtension, FlutterDevExtension])
def extension_class(request):
    return request.param


def test_extension(extension_class):
    flutter_extension = extension_class(
        extension_name="flutter", yaml_data=dict(base="core18")
    )

    assert flutter_extension.root_snippet == {
        "assumes": ["snapd2.43"],
        "plugs": {
            "gtk-3-themes": {
                "interface": "content",
                "target": "$SNAP/data-dir/themes",
                "default-provider": "gtk-common-themes",
            },
            "icon-themes": {
                "interface": "content",
                "target": "$SNAP/data-dir/icons",
                "default-provider": "gtk-common-themes",
            },
            "sound-themes": {
                "interface": "content",
                "target": "$SNAP/data-dir/sounds",
                "default-provider": "gtk-common-themes",
            },
            "gnome-3-28-1804": {
                "interface": "content",
                "target": "$SNAP/gnome-platform",
                "default-provider": "gnome-3-28-1804",
            },
        },
        "environment": {"SNAP_DESKTOP_RUNTIME": "$SNAP/gnome-platform"},
        "layout": {
            "/usr/share/xml/iso-codes": {
                "bind": "$SNAP/gnome-platform/usr/share/xml/iso-codes"
            }
        },
    }

    assert flutter_extension.app_snippet == {
        "command-chain": ["snap/command-chain/desktop-launch"],
        "plugs": ["desktop", "desktop-legacy", "gsettings", "opengl", "wayland", "x11"],
    }

    assert flutter_extension.part_snippet == dict()

    # TODO Hack until we move extension name to each extension class.
    channel = "dev" if "dev" in flutter_extension.__repr__() else "master"
    assert flutter_extension.parts == {
        "gnome-3-28-extension": {
            "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
            "source-subdir": "gnome",
            "plugin": "make",
            "make-parameters": ["PLATFORM_PLUG=gnome-3-28-1804"],
            "build-packages": ["gcc", "libgtk-3-dev"],
        },
        "flutter-extension": {
            "plugin": "nil",
            "override-pull": textwrap.dedent(
                f"""\
                            flutter channel {channel}
                            flutter config --enable-linux-desktop
                            flutter upgrade
                            flutter doctor
                            """
            ),
            "build-snaps": ["flutter/latest/stable"],
        },
    }


def test_supported_bases(extension_class):
    assert extension_class.get_supported_bases() == ("core18",)


def test_supported_confinement(extension_class):
    extension_class.get_supported_confinement() == ("strict", "devmode")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from simplejson.scanner import JSONDecodeError
from testtools.matchers import Contains, Equals, Not

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class RegisterTestCase(FakeStoreCommandsBaseTestCase):
    def test_register_without_name_must_error(self):
        result = self.run_command(["register"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_register_without_login_must_ask(self):
        self.fake_store_register.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            None,
        ]

        result = self.run_command(
            ["register", "snap-name"], input="y\nuser@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_register_name_successfully(self):
        result = self.run_command(["register", "test-snap"], input="y\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Registering test-snap"))
        self.assertThat(
            result.output,
            Contains("Congrats! You are now the publisher of 'test-snap'."),
        )
        self.assertThat(
            result.output,
            Not(Contains("Congratulations! You're now the publisher for 'test-snap'.")),
        )
        self.fake_store_register.mock.assert_called_once_with(
            "test-snap", is_private=False, series="16", store_id=None
        )

    def test_register_name_to_specific_store_successfully(self):
        result = self.run_command(
            ["register", "test-snap", "--store", "my-brand"], input="y\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Registering test-snap"))
        self.assertThat(
            result.output,
            Contains("Congrats! You are now the publisher of 'test-snap'."),
        )
        self.assertThat(
            result.output,
            Not(Contains("Congratulations! You're now the publisher for 'test-snap'.")),
        )
        self.fake_store_register.mock.assert_called_once_with(
            "test-snap", is_private=False, series="16", store_id="my-brand"
        )

    def test_register_private_name_successfully(self):
        result = self.run_command(["register", "test-snap", "--private"], input="y\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains("Even though this is private snap, you should think carefully"),
        )
        self.assertThat(result.output, Contains("Registering test-snap"))
        self.assertThat(
            result.output,
            Contains("Congrats! You are now the publisher of 'test-snap'."),
        )
        self.assertThat(
            result.output,
            Not(Contains("Congratulations! You're now the publisher for 'test-snap'.")),
        )
        self.fake_store_register.mock.assert_called_once_with(
            "test-snap", is_private=True, series="16", store_id=None
        )

    def test_registration_failed(self):
        response = mock.Mock()
        response.json.side_effect = JSONDecodeError("mock-fail", "doc", 1)
        self.fake_store_register.mock.side_effect = storeapi.errors.StoreRegistrationError(
            "test-snap", response
        )

        raised = self.assertRaises(
            storeapi.errors.StoreRegistrationError,
            self.run_command,
            ["register", "test-snap"],
            input="y\n",
        )

        self.assertThat(str(raised), Equals("Registration failed."))

    def test_registration_cancelled(self):
        response = mock.Mock()
        response.json.side_effect = JSONDecodeError("mock-fail", "doc", 1)
        self.fake_store_register.mock.side_effect = storeapi.errors.StoreRegistrationError(
            "test-snap", response
        )

        result = self.run_command(["register", "test-snap"], input="n\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Thank you! 'test-snap' will remain available")
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

import click

from snapcraft.internal import review_tools

from . import echo


def review_snap(*, snap_file: str) -> None:
    # Review the snap.
    if review_tools.is_available():
        echo.info(
            "Running the review tools before pushing this snap to the Snap " "Store."
        )
        # TODO just raise when we can override.
        try:
            review_tools.run(snap_filename=snap_file)
        except review_tools.errors.ReviewError as review_error:
            echo.warning(review_error.get_brief())
            echo.warning(review_error.get_resolution())
            # This can get too colorful if we don't just print it.
            click.echo(review_error.get_details())
    elif sys.platform == "linux":
        echo.warning(
            "Install the review-tools from the Snap Store for enhanced checks "
            "before uploading this snap."
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from pathlib import Path

from testtools.matchers import Equals, FileContains, Is

from snapcraft.internal.meta._snap_packaging import _SnapPackaging
from snapcraft.internal.project_loader import load_config
from snapcraft.project import Project
from tests import fixture_setup, unit


class SnapPackagingRunnerTests(unit.TestCase):
    def _get_snap_packaging(self, **yaml_args):
        if "parts" not in yaml_args:
            yaml_args["parts"] = dict(part1=dict(plugin="nil"))

        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, **yaml_args)
        self.useFixture(snapcraft_yaml)

        project = Project(
            snapcraft_yaml_file_path=snapcraft_yaml.snapcraft_yaml_file_path
        )
        config = load_config(project)

        return _SnapPackaging(project_config=config, extracted_metadata=None)

    def test_no_apps(self):
        apps = dict()

        sp = self._get_snap_packaging(apps=apps, confinement="strict")
        runner = sp._generate_snapcraft_runner()

        self.assertThat(runner, Equals(None))

        sp = self._get_snap_packaging(apps=apps, confinement="strict")
        runner = sp._generate_snapcraft_runner()

        self.assertThat(runner, Equals(None))

    def test_strict_app(self):
        apps = dict(testapp=dict(command="echo"))

        sp = self._get_snap_packaging(apps=apps, confinement="strict")
        runner = sp._generate_snapcraft_runner()

        self.assertThat(runner, Equals("snap/command-chain/snapcraft-runner"))

        runner_path = Path(self.path, "prime", runner)

        with open(runner_path, "r") as f:
            snapcraft_runner = f.read()

        expected_runner = textwrap.dedent(
            """
            #!/bin/sh
            export PATH="$SNAP/usr/sbin:$SNAP/usr/bin:$SNAP/sbin:$SNAP/bin${PATH:+:$PATH}"
            export LD_LIBRARY_PATH="$SNAP_LIBRARY_PATH${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
            exec "$@"
            """
        ).lstrip()

        self.assertThat(snapcraft_runner, Equals(expected_runner))

    def test_classic_app_with_snapd_workaround(self):
        """Test workaround for classic (LP: #1860369)."""
        apps = dict(testapp=dict(command="echo"))

        for base in ("core", "core16", "core18"):
            sp = self._get_snap_packaging(apps=apps, base=base, confinement="classic")
            runner = sp._generate_snapcraft_runner()

            self.expectThat(runner, Equals("snap/command-chain/snapcraft-runner"))
            runner_path = Path(self.path, "prime", runner)
            self.expectThat(
                runner_path,
                FileContains(
                    textwrap.dedent(
                        """
            #!/bin/sh

            exec "$@"
            """
                    ).lstrip()
                ),
            )

    def test_classic_app_without_snapd_workaround(self):
        """Test no workaround for classic (LP: #1860369)."""
        apps = dict(testapp=dict(command="echo"))

        sp = self._get_snap_packaging(apps=apps, base="core20", confinement="classic")
        runner = sp._generate_snapcraft_runner()

        self.expectThat(runner, Is(None))

    def test_snapd(self):
        apps = dict(testapp=dict(command="echo"))

        sp = self._get_snap_packaging(
            apps=apps,
            confinement="classic",
            type="snapd",
            base=None,
            build_base="core18",
        )
        runner = sp._generate_snapcraft_runner()

        self.assertThat(runner, Equals(None))

    def test_assembled_runtime_environment_classic(self):
        apps = dict(testapp=dict(command="echo"))

        sp = self._get_snap_packaging(
            apps=apps, confinement="classic", type="app", base="core"
        )

        assembled_env = sp._assemble_runtime_environment()

        self.assertThat(assembled_env, Equals(""))

    def test_assembled_runtime_environment_strict(self):
        apps = dict(testapp=dict(command="echo"))

        sp = self._get_snap_packaging(
            apps=apps, confinement="strict", type="app", base="core"
        )

        assembled_env = sp._assemble_runtime_environment()

        expected_env = textwrap.dedent(
            """
            export PATH="$SNAP/usr/sbin:$SNAP/usr/bin:$SNAP/sbin:$SNAP/bin${PATH:+:$PATH}"
            export LD_LIBRARY_PATH="$SNAP_LIBRARY_PATH${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
            """
        ).strip()

        self.assertThat(assembled_env, Equals(expected_env))

    def test_assembled_runtime_environment_strict_patched(self):
        apps = dict(testapp=dict(command="echo"))
        parts = {"part1": {"plugin": "nil", "build-attributes": ["enable-patchelf"]}}

        sp = self._get_snap_packaging(
            apps=apps, parts=parts, confinement="strict", type="app", base="core"
        )

        assembled_env = sp._assemble_runtime_environment()

        # Verify that, since all parts are using patchelf, no LD_LIBRARY_PATH is set
        expected_env = textwrap.dedent(
            """
            export PATH="$SNAP/usr/sbin:$SNAP/usr/bin:$SNAP/sbin:$SNAP/bin${PATH:+:$PATH}"
            """
        ).strip()

        self.assertThat(assembled_env, Equals(expected_env))

    def test_stub_hook(self):
        apps = dict(testapp=dict(command="echo"))
        parts = {"part1": {"plugin": "nil", "build-attributes": ["enable-patchelf"]}}
        hooks = dict(configure={"command-chain": ["foo"], "plugs": ["network"]})

        sp = self._get_snap_packaging(
            apps=apps,
            parts=parts,
            confinement="strict",
            type="app",
            base="core",
            hooks=hooks,
        )

        sp.generate_hook_wrappers()

        hook = Path(sp._prime_dir, "meta", "hooks", "configure")
        assert hook.exists()
        assert hook.read_text() == "#!/bin/sh\n"
        assert hook.stat().st_mode & 0xFFF == 0o755

    def test_no_stub_for_preconfigured_hook(self):
        apps = dict(testapp=dict(command="echo"))
        parts = {"part1": {"plugin": "nil", "build-attributes": ["enable-patchelf"]}}
        hooks = dict(configure={"command-chain": ["foo"], "plugs": ["network"]})

        sp = self._get_snap_packaging(
            apps=apps,
            parts=parts,
            confinement="strict",
            type="app",
            base="core",
            hooks=hooks,
        )

        hook = Path(sp._prime_dir, "meta", "hooks", "configure")
        hook.parent.mkdir(parents=True)
        hook.write_text("foo")

        sp.generate_hook_wrappers()

        assert hook.exists()
        assert hook.read_text() == "foo"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.plugins.v2.meson import MesonPlugin


class MesonPluginTest(TestCase):
    def test_schema(self):
        schema = MesonPlugin.get_schema()

        self.assertThat(
            schema,
            Equals(
                {
                    "$schema": "http://json-schema.org/draft-04/schema#",
                    "additionalProperties": False,
                    "properties": {
                        "meson-parameters": {
                            "default": [],
                            "items": {"type": "string"},
                            "type": "array",
                            "uniqueItems": True,
                        },
                        "meson-version": {"default": "", "type": "string"},
                    },
                    "required": ["source"],
                    "type": "object",
                }
            ),
        )

    def test_get_build_packages(self):
        plugin = MesonPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(
            plugin.get_build_packages(),
            Equals(
                {
                    "ninja-build",
                    "gcc",
                    "python3-pip",
                    "python3-setuptools",
                    "python3-wheel",
                }
            ),
        )

    def test_get_build_environment(self):
        plugin = MesonPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_environment(), Equals(dict()))

    def test_get_build_commands(self):
        class Options:
            meson_parameters = list()
            meson_version = ""

        plugin = MesonPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    "/usr/bin/python3 -m pip install -U meson",
                    '[ ! -f build.ninja ] && meson "${SNAPCRAFT_PART_SRC_WORK}"',
                    "ninja",
                    'DESTDIR="${SNAPCRAFT_PART_INSTALL}" ninja install',
                ]
            ),
        )

    def test_get_build_commands_with_options(self):
        class Options:
            meson_parameters = ["--buildtype=release"]
            meson_version = "2.2"

        plugin = MesonPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    "/usr/bin/python3 -m pip install -U meson==2.2",
                    '[ ! -f build.ninja ] && meson --buildtype=release "${SNAPCRAFT_PART_SRC_WORK}"',
                    "ninja",
                    'DESTDIR="${SNAPCRAFT_PART_INSTALL}" ninja install',
                ]
            ),
        )

    def test_out_of_source_build(self):
        plugin = MesonPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.out_of_source_build, Equals(True))



import launch
import launch_ros.actions


def generate_launch_description():
    exit_after_receive = launch.actions.DeclareLaunchArgument(
        "exit_after_receive",
        default_value="false",
        description="Exit after receiving one message.",
    )

    talker = launch_ros.actions.Node(
        package="talker_cpp",
        node_executable="talker",
        node_name="talker",
        output="screen",
    )

    listener = launch_ros.actions.Node(
        package="listener_py",
        node_executable="listener",
        node_name="listener",
        output="screen",
        on_exit=launch.actions.Shutdown(),
        parameters=[
            {
                "exit_after_receive": launch.substitutions.LaunchConfiguration(
                    "exit_after_receive"
                )
            }
        ],
    )

    return launch.LaunchDescription([exit_after_receive, talker, listener])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import platform
from unittest import mock

import testtools
from testtools.matchers import Equals

import snapcraft
from snapcraft.internal import common
from snapcraft.internal.errors import SnapcraftEnvironmentError
from snapcraft.project._project_options import (
    _32BIT_USERSPACE_ARCHITECTURE,
    _get_platform_architecture,
)
from tests import unit


class TestNativeOptions:

    scenarios = [
        (
            "amd64",
            dict(
                machine="x86_64",
                architecture=("64bit", "ELF"),
                expected_arch_triplet="x86_64-linux-gnu",
                expected_deb_arch="amd64",
                expected_kernel_arch="x86",
                expected_core_dynamic_linker="lib64/ld-linux-x86-64.so.2",
            ),
        ),
        (
            "amd64-kernel-i686-userspace",
            dict(
                machine="x86_64",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="i386-linux-gnu",
                expected_deb_arch="i386",
                expected_kernel_arch="x86",
                expected_core_dynamic_linker="lib/ld-linux.so.2",
            ),
        ),
        (
            "i686",
            dict(
                machine="i686",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="i386-linux-gnu",
                expected_deb_arch="i386",
                expected_kernel_arch="x86",
                expected_core_dynamic_linker="lib/ld-linux.so.2",
            ),
        ),
        (
            "armv7l",
            dict(
                machine="armv7l",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="arm-linux-gnueabihf",
                expected_deb_arch="armhf",
                expected_kernel_arch="arm",
                expected_core_dynamic_linker="lib/ld-linux-armhf.so.3",
            ),
        ),
        (
            "aarch64",
            dict(
                machine="aarch64",
                architecture=("64bit", "ELF"),
                expected_arch_triplet="aarch64-linux-gnu",
                expected_deb_arch="arm64",
                expected_kernel_arch="arm64",
                expected_core_dynamic_linker="lib/ld-linux-aarch64.so.1",
            ),
        ),
        (
            "aarch64-kernel-armv7l-userspace",
            dict(
                machine="aarch64",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="arm-linux-gnueabihf",
                expected_deb_arch="armhf",
                expected_kernel_arch="arm",
                expected_core_dynamic_linker="lib/ld-linux-armhf.so.3",
            ),
        ),
        (
            "armv8l-kernel-armv7l-userspace",
            dict(
                machine="armv8l",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="arm-linux-gnueabihf",
                expected_deb_arch="armhf",
                expected_kernel_arch="arm",
                expected_core_dynamic_linker="lib/ld-linux-armhf.so.3",
            ),
        ),
        (
            "ppc",
            dict(
                machine="ppc",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="powerpc-linux-gnu",
                expected_deb_arch="powerpc",
                expected_kernel_arch="powerpc",
                expected_core_dynamic_linker="lib/ld-linux.so.2",
            ),
        ),
        (
            "ppc64le",
            dict(
                machine="ppc64le",
                architecture=("64bit", "ELF"),
                expected_arch_triplet="powerpc64le-linux-gnu",
                expected_deb_arch="ppc64el",
                expected_kernel_arch="powerpc",
                expected_core_dynamic_linker="lib64/ld64.so.2",
            ),
        ),
        (
            "ppc64le-kernel-ppc-userspace",
            dict(
                machine="ppc64le",
                architecture=("32bit", "ELF"),
                expected_arch_triplet="powerpc-linux-gnu",
                expected_deb_arch="powerpc",
                expected_kernel_arch="powerpc",
                expected_core_dynamic_linker="lib/ld-linux.so.2",
            ),
        ),
        (
            "riscv64",
            dict(
                machine="riscv64",
                architecture=("64bit", "ELF"),
                expected_arch_triplet="riscv64-linux-gnu",
                expected_deb_arch="riscv64",
                expected_kernel_arch="riscv64",
                expected_core_dynamic_linker="lib/ld-linux-riscv64-lp64d.so.1",
            ),
        ),
        (
            "s390x",
            dict(
                machine="s390x",
                architecture=("64bit", "ELF"),
                expected_arch_triplet="s390x-linux-gnu",
                expected_deb_arch="s390x",
                expected_kernel_arch="s390",
                expected_core_dynamic_linker="lib/ld64.so.1",
            ),
        ),
    ]

    def test_architecture_options(
        self,
        monkeypatch,
        machine,
        architecture,
        expected_arch_triplet,
        expected_deb_arch,
        expected_kernel_arch,
        expected_core_dynamic_linker,
    ):
        monkeypatch.setattr(platform, "architecture", lambda: architecture)
        monkeypatch.setattr(platform, "machine", lambda: machine)

        options = snapcraft.ProjectOptions()

        assert options.arch_triplet == expected_arch_triplet
        assert options.deb_arch == expected_deb_arch
        assert options.kernel_arch == expected_kernel_arch

        # The core dynamic linker is correct.  Guard against stray absolute
        # paths, as they cause os.path.join to discard the previous
        # argument.
        monkeypatch.setattr(os.path, "lexists", lambda x: True)
        monkeypatch.setattr(os.path, "islink", lambda x: False)
        expected_linker_path = os.path.join(
            common.get_installed_snap_path("core"), expected_core_dynamic_linker
        )
        assert options.get_core_dynamic_linker("core") == expected_linker_path

    def test_get_platform_architecture(
        self,
        monkeypatch,
        machine,
        architecture,
        expected_arch_triplet,
        expected_deb_arch,
        expected_kernel_arch,
        expected_core_dynamic_linker,
    ):
        monkeypatch.setattr(platform, "architecture", lambda: architecture)
        monkeypatch.setattr(platform, "machine", lambda: machine)

        platform_arch = _get_platform_architecture()
        userspace_conversions = _32BIT_USERSPACE_ARCHITECTURE

        if architecture[0] == "32bit" and machine in userspace_conversions:
            assert platform_arch == userspace_conversions[machine]
        else:
            assert platform_arch == machine


class OptionsTestCase(unit.TestCase):
    def test_cross_compiler_prefix_missing(self):
        options = snapcraft.ProjectOptions(target_deb_arch="x86_64")

        with testtools.ExpectedException(
            SnapcraftEnvironmentError,
            "Cross compilation not supported for target arch 'x86_64'",
        ):
            options.cross_compiler_prefix

    @mock.patch("platform.architecture")
    @mock.patch("platform.machine")
    def test_cross_compiler_prefix_empty(
        self, mock_platform_machine, mock_platform_architecture
    ):
        mock_platform_machine.return_value = "x86_64"
        mock_platform_architecture.return_value = ("64bit", "ELF")
        options = snapcraft.ProjectOptions(target_deb_arch="i386")
        self.assertThat(options.cross_compiler_prefix, Equals(""))


class TestHostIsCompatibleWithTargetBase:

    scenarios = (
        ("trusty core", dict(codename="trusty", base="core", is_compatible=True)),
        ("xenial core", dict(codename="xenial", base="core", is_compatible=True)),
        ("bionic core", dict(codename="bionic", base="core", is_compatible=False)),
        ("trusty core18", dict(codename="trusty", base="core18", is_compatible=True)),
        ("xenial core18", dict(codename="xenial", base="core18", is_compatible=True)),
        ("bionic core18", dict(codename="bionic", base="core18", is_compatible=True)),
        (
            "Random codename core18",
            dict(codename="random", base="core18", is_compatible=False),
        ),
        (
            "trusty unknown-base",
            dict(codename="trusty", base="unknown", is_compatible=False),
        ),
    )

    def test_compatibility(self, monkeypatch, codename, base, is_compatible):
        monkeypatch.setattr(
            snapcraft.internal.os_release.OsRelease,
            "version_codename",
            lambda x: codename,
        )

        assert (
            snapcraft.ProjectOptions().is_host_compatible_with_base(base)
            is is_compatible
        )


class TestLinkerVersionForBase(unit.TestCase):
    def setUp(self):
        super().setUp()
        patcher = mock.patch("snapcraft.file_utils.get_linker_version_from_file")
        self.get_linker_version_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_get_linker_version_for_core(self):
        self.assertThat(
            snapcraft.ProjectOptions()._get_linker_version_for_base("core"),
            Equals("2.23"),
        )
        self.get_linker_version_mock.assert_not_called()

    def test_get_linker_version_for_core18(self):
        self.assertThat(
            snapcraft.ProjectOptions()._get_linker_version_for_base("core18"),
            Equals("2.27"),
        )
        self.get_linker_version_mock.assert_not_called()

    def test_get_linker_version_for_random_core(self):
        self.get_linker_version_mock.return_value = "4.10"
        self.assertThat(
            snapcraft.ProjectOptions()._get_linker_version_for_base("random"),
            Equals("4.10"),
        )
        self.get_linker_version_mock.assert_called_once_with("ld-2.23.so")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tarfile
from collections import OrderedDict
from pathlib import Path

from testtools.matchers import Equals

from snapcraft import yaml_utils
from snapcraft.internal.remote_build import WorkTree
from snapcraft.project import Project
from tests import fixture_setup, unit

from . import TestDir


class WorkTreeTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self._source = self.useFixture(TestDir())
        self._source.create_file("foo")
        self._source.create_file("bar")
        self._source.create_dir("dir")
        self._source.create_file("dir", "baz")
        self._source.create_dir("empty_dir")

        self._dest = self.useFixture(TestDir())

        self._snapcraft_yaml = fixture_setup.SnapcraftYaml(self._source.path)
        self._snapcraft_yaml.update_part(
            "my-part", {"plugin": "nil", "source": self._source.path}
        )
        self.useFixture(self._snapcraft_yaml)
        self.load_project_and_worktree()

    def load_project_and_worktree(self):
        """(Re)load project and worktree."""
        self._project = Project(
            snapcraft_yaml_file_path=self._snapcraft_yaml.snapcraft_yaml_file_path
        )
        self._project._project_dir = self._source.path
        self._wt = WorkTree(self._dest.path, self._project)

    def archive_path(self, part_name, source_selector=None):
        if source_selector:
            return os.path.join(
                self._dest.path,
                "repo",
                "sources",
                part_name,
                source_selector,
                part_name + ".tar.gz",
            )
        return os.path.join(
            self._dest.path, "repo", "sources", part_name, part_name + ".tar.gz"
        )

    def tarball_file_list(self, part_name, source_selector):
        archive_path = self.archive_path(part_name, source_selector)
        file_list = []
        with tarfile.open(archive_path, "r") as t:
            for m in t.getmembers():
                file_list.append(os.path.relpath(m.name, "./"))
        return file_list

    def tarball_file_contains(self, part_name, source_selector, *paths):
        file_list = self.tarball_file_list(part_name, source_selector)
        return os.path.join(*paths) in file_list

    def test_worktree_creation(self):
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache"))
        self.assertTrue(self._dest.exists("cache", "my-part"))
        self.assertTrue(self._dest.exists("repo"))
        self.assertTrue(self._dest.exists("repo", "sources"))
        self.assertTrue(self._dest.exists("repo", "sources", "my-part"))
        self.assertTrue(
            self._dest.exists("repo", "sources", "my-part", "my-part.tar.gz")
        )
        self.assertTrue(self._dest.exists("repo", "snap"))
        self.assertTrue(self._dest.exists("repo", "snap", "snapcraft.yaml"))

    def test_worktree_cache_new_file(self):
        self._wt.prepare_repository()
        self._source.create_file("new_file")
        self.assertFalse(self._dest.exists("cache", "my-part", "new_file"))
        self.assertFalse(self.tarball_file_contains("my-part", None, "new_file"))
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache", "my-part", "new_file"))
        self.assertTrue(self.tarball_file_contains("my-part", None, "new_file"))

    def test_worktree_cache_remove_file(self):
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache", "my-part", "foo"))
        self.assertTrue(self.tarball_file_contains("my-part", None, "foo"))
        self._source.unlink("foo")
        self._wt.prepare_repository()
        self.assertFalse(self._dest.exists("cache", "my-part", "foo"))
        self.assertFalse(self.tarball_file_contains("my-part", None, "foo"))

    def test_worktree_cache_remove_empty_dir(self):
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache", "my-part", "empty_dir"))
        self.assertTrue(self.tarball_file_contains("my-part", None, "empty_dir"))
        self._source.rmdir("empty_dir")
        self._wt.prepare_repository()
        self.assertFalse(self._dest.exists("cache", "my-part", "empty_dir"))
        self.assertFalse(self.tarball_file_contains("my-part", None, "empty_dir"))

    def test_worktree_cache_remove_dir(self):
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache", "my-part", "dir"))
        self.assertTrue(self.tarball_file_contains("my-part", None, "dir"))
        self._source.unlink("dir", "baz")
        self._source.rmdir("dir")
        self._wt.prepare_repository()
        self.assertFalse(self._dest.exists("cache", "my-part", "dir"))
        self.assertFalse(self.tarball_file_contains("my-part", None, "dir"))

    def test_assets_yaml(self):
        self._source.create_dir("snap", "gui")
        self._source.create_file("snap", "gui", "test.desktop")
        self._source.create_file("snap", "gui", "test.png")
        self.assertFalse(self._dest.exists("repo", "snap", "gui", "test.desktop"))
        self.assertFalse(self._dest.exists("repo", "snap", "gui", "test.png"))
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("repo", "snap", "gui", "test.desktop"))
        self.assertTrue(self._dest.exists("repo", "snap", "gui", "test.png"))

    def test_deprecated_local_plugins(self):
        self._source.create_dir("snap", "plugins")
        self._source.create_file("snap", "plugins", "plugin.py")
        self.assertFalse(self._dest.exists("repo", "snap", "plugins", "plugin.py"))
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("repo", "snap", "plugins", "plugin.py"))

    def test_source_selectors(self):
        self._snapcraft_yaml.update_part(
            "src-selectors",
            {
                "plugin": "nil",
                "source": [
                    {"on i386": self._source.path},
                    {"on arm64": self._source.path},
                ],
            },
        )
        self.load_project_and_worktree()
        self._wt.prepare_repository()
        self.assertTrue(self._dest.exists("cache"))
        self.assertTrue(self._dest.exists("cache", "src-selectors"))
        self.assertTrue(self._dest.exists("cache", "src-selectors", "on i386"))
        self.assertTrue(self._dest.exists("cache", "src-selectors", "on arm64"))
        self.assertTrue(self._dest.exists("repo"))
        self.assertTrue(self._dest.exists("repo", "sources"))
        self.assertTrue(self._dest.exists("repo", "sources", "my-part"))
        self.assertTrue(
            self._dest.exists("repo", "sources", "my-part", "my-part.tar.gz")
        )
        self.assertTrue(
            self._dest.exists(
                "repo", "sources", "src-selectors", "on i386", "src-selectors.tar.gz"
            )
        )
        self.assertTrue(
            self._dest.exists(
                "repo", "sources", "src-selectors", "on arm64", "src-selectors.tar.gz"
            )
        )
        self.assertFalse(
            self._dest.exists(
                "repo", "sources", "src-selectors", "src-selectors.tar.gz"
            )
        )
        self.assertTrue(self._dest.exists("repo", "snap"))
        self.assertTrue(self._dest.exists("repo", "snap", "snapcraft.yaml"))

    def test_stripped_source_keys(self):
        self._snapcraft_yaml.update_part(
            "my-part",
            {
                "plugin": "nil",
                "source": self._source.path,
                "source-checksum": "strip-me",
                "source-branch": "strip-me",
                "source-commit": "strip-me",
                "source-depth": "strip-me",
                "source-subdir": "test-sub-dir",
                "source-tag": "strip-me",
                "source-type": "local",
            },
        )

        self.load_project_and_worktree()
        self._wt.prepare_repository()

        config = yaml_utils.load_yaml_file(
            str(Path(self._dest.path, "repo", "snap", "snapcraft.yaml"))
        )

        self.assertThat(
            config["parts"]["my-part"],
            Equals(
                OrderedDict(
                    [
                        ("plugin", "nil"),
                        ("source", "sources/my-part/my-part.tar.gz"),
                        ("source-subdir", "test-sub-dir"),
                        ("source-type", "tar"),
                    ]
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import textwrap

import yaml as yaml_utils
from testtools.matchers import Equals, FileExists

import tests
from tests import fixture_setup, integration


class AppstreamMetadataTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        with open("test.metainfo.xml", "w") as appstream_file:
            appstream_file.write(
                textwrap.dedent(
                    """<?xml version="1.0" encoding="UTF-8"?>
                <component>
                  <id>com.example.testid</id>
                  <description>test-appstream-description</description>
                  <summary>test-appstream-summary</summary>
                  <icon type="local">assets/test-icon.png</icon>
                  <launchable type="desktop-id">
                    com.example.test-app.desktop
                  </launchable>
                </component>"""
                )
            )

    def test_metadata_extracted_from_appstream(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, summary=None, description=None
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part",
            {"plugin": "dump", "source": ".", "parse-info": ["test.metainfo.xml"]},
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")
        with open(os.path.join("prime", "meta", "snap.yaml")) as snap_yaml_file:
            snap_yaml = yaml_utils.load(snap_yaml_file)
        self.assertThat(snap_yaml["description"], Equals("test-appstream-description"))
        self.assertThat(snap_yaml["summary"], Equals("test-appstream-summary"))

    def test_specified_metadata_not_overwritten(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path, description=None)
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part",
            {"plugin": "dump", "source": ".", "parse-info": ["test.metainfo.xml"]},
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")
        with open(os.path.join("prime", "meta", "snap.yaml")) as snap_yaml_file:
            snap_yaml = yaml_utils.load(snap_yaml_file)
        self.assertThat(snap_yaml["description"], Equals("test-appstream-description"))
        self.assertThat(snap_yaml["summary"], Equals("test-summary"))

    def test_icon_extracted_from_appstream(self):
        os.mkdir("assets")
        shutil.copyfile(
            os.path.join(os.path.dirname(tests.__file__), "data", "icon.png"),
            os.path.join(self.path, "assets", "test-icon.png"),
        )

        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, summary=None, description=None
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part",
            {"plugin": "dump", "source": ".", "parse-info": ["test.metainfo.xml"]},
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        self.assertThat(os.path.join("prime", "meta", "gui", "icon.png"), FileExists())

    def test_desktop_extracted_from_appstream(self):
        xdg_data_dir = os.path.join("usr", "local", "share", "applications")
        os.makedirs(os.path.join(xdg_data_dir, "com.example.test"))
        shutil.copyfile(
            os.path.join(os.path.dirname(tests.__file__), "data", "test.desktop"),
            os.path.join(self.path, xdg_data_dir, "com.example.test", "app.desktop"),
        )

        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path, summary=None, description=None
        )
        snapcraft_yaml.data["adopt-info"] = "test-part"
        snapcraft_yaml.update_part(
            "test-part",
            {"plugin": "dump", "source": ".", "parse-info": ["test.metainfo.xml"]},
        )
        snapcraft_yaml.update_app(
            "test-app", {"command": "echo", "common-id": "com.example.testid"}
        )
        self.useFixture(snapcraft_yaml)

        self.run_snapcraft("prime")

        self.assertThat(
            os.path.join("prime", "meta", "gui", "test-app.desktop"), FileExists()
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
from unittest import mock

import pytest
from testtools import TestCase
from testtools.matchers import Contains, Equals

# required for schema format checkers
import snapcraft.internal.project_loader._config  # noqa: F401
import snapcraft.yaml_utils.errors
from snapcraft.project._schema import Validator

from . import ProjectBaseTest


def get_data():
    return {
        "name": "my-package-1",
        "base": "core18",
        "version": "1.0-snapcraft1~ppa1",
        "summary": "my summary less that 79 chars",
        "description": "description which can be pretty long",
        "adopt-info": "part1",
        "parts": {"part1": {"plugin": "project", "parse-info": ["test-metadata-file"]}},
    }


@pytest.fixture
def data():
    """Return snapcraft.yaml data to validate."""
    return get_data()


class ValidationBaseTest(TestCase):
    def setUp(self):
        super().setUp()

        self.data = get_data()


class ValidationTest(ValidationBaseTest):
    def test_summary_too_long(self):
        self.data["summary"] = "a" * 80
        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            Validator(self.data).validate,
        )

        expected_message = (
            "The 'summary' property does not match the required schema: "
            "'{}' is too long (maximum length is 78)"
        ).format(self.data["summary"])
        self.assertThat(raised.message, Equals(expected_message), message=self.data)

    def test_apps_required_properties(self):
        self.data["apps"] = {"service1": {}}

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            Validator(self.data).validate,
        )

        expected_message = (
            "The 'apps/service1' property does not match the "
            "required schema: 'command' is a required "
            "property"
        )
        self.assertThat(raised.message, Equals(expected_message), message=self.data)

    def test_schema_file_not_found(self):
        mock_the_open = mock.mock_open()
        mock_the_open.side_effect = FileNotFoundError()

        with mock.patch("snapcraft.project._schema.open", mock_the_open, create=True):
            raised = self.assertRaises(
                snapcraft.yaml_utils.errors.YamlValidationError, Validator, self.data
            )

        expected_message = "snapcraft validation file is missing from installation path"
        self.assertThat(raised.message, Equals(expected_message))

    def test_valid_app_daemons(self):
        self.data["apps"] = {
            "service1": {"command": "binary1 start", "daemon": "simple"},
            "service2": {
                "command": "binary2",
                "stop-command": "binary2 --stop",
                "daemon": "simple",
            },
            "service3": {"command": "binary3", "daemon": "forking"},
            "service4": {
                "command": "binary4",
                "daemon": "simple",
                "restart-condition": "always",
            },
            "service5": {"command": "binary5", "daemon": "notify"},
            "service6": {
                "command": "binary6",
                "post-stop-command": "binary6 --post-stop",
                "daemon": "simple",
            },
            "service7": {
                "command": "binary7",
                "reload-command": "binary7 --reload",
                "daemon": "simple",
            },
            "service8": {
                "command": "binary8",
                "daemon": "dbus",
                "bus-name": "org.test.snapcraft",
            },
            "service9": {
                "command": "binary9",
                "daemon": "simple",
                "start-timeout": "1s",
            },
            "service10": {
                "command": "binary10",
                "daemon": "simple",
                "stop-timeout": "1s",
            },
            "service11": {
                "command": "binary11",
                "daemon": "simple",
                "restart-delay": "1s",
            },
            "service12": {
                "command": "binary12",
                "daemon": "simple",
                "watchdog-timeout": "1s",
            },
            "service13": {
                "command": "binary13",
                "daemon": "oneshot",
                "timer": "mon,10:00-12:00",
            },
            "service14": {
                "command": "binary14",
                "daemon": "simple",
                "restart-condition": "on-watchdog",
                "watchdog-timeout": "30s",
            },
            "service15": {
                "command": "binary15",
                "daemon": "simple",
                "install-mode": "enable",
            },
            "service16": {
                "command": "binary16",
                "daemon": "simple",
                "install-mode": "disable",
            },
        }

        Validator(self.data).validate()

    def test_invalid_restart_condition(self):
        self.data["apps"] = {
            "service1": {
                "command": "binary1",
                "daemon": "simple",
                "restart-condition": "on-tuesday",
            }
        }

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            Validator(self.data).validate,
        )

        self.assertThat(
            str(raised),
            Contains(
                "The 'apps/service1/restart-condition' property does not match "
                "the required schema: 'on-tuesday' is not one of ['on-success', "
                "'on-failure', 'on-abnormal', 'on-abort', 'on-watchdog', 'always', "
                "'never']"
            ),
        )

    def test_missing_required_property_and_missing_adopt_info(self):
        del self.data["summary"]
        del self.data["adopt-info"]

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            Validator(self.data).validate,
        )

        expected_message = (
            "'adopt-info' is a required property or 'summary' is a required property"
        )
        self.assertThat(raised.message, Equals(expected_message), message=self.data)

    def test_invalid_install_mode(self):
        self.data["apps"] = {
            "service": {
                "command": "binary",
                "daemon": "simple",
                "install-mode": "invalid",
            }
        }

        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            Validator(self.data).validate,
        )

        expected_message = (
            "The 'apps/service/install-mode' property does not match the required schema: "
            "'invalid' is not one of ['enable', 'disable']"
        )
        self.assertThat(raised.message, Equals(expected_message), message=self.data)


@pytest.mark.parametrize(
    "option,value",
    [
        ("stop-command", "binary1 --stop"),
        ("post-stop-command", "binary1 --post-stop"),
        ("before", ["service1"]),
        ("after", ["service2"]),
    ],
)
def test_daemon_dependency(data, option, value):
    data["apps"] = {"service1": {"command": "binary1", option: value}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    assert str(error.value).endswith(
        "The 'apps/service1' property does not match the required schema: "
        f"'daemon' is a dependency of {option!r}"
    )


@pytest.mark.parametrize("key", ["name", "parts"])
def test_required_properties(data, key):
    del data[key]

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    assert f"{key!r} is a required property" in str(error.value)


class TestInvalidNames:
    e1 = "not a valid snap name. Snap names can only use ASCII lowercase letters, numbers, and hyphens, and must have at least one letter."
    e2 = "not a valid snap name. Snap names cannot start with a hyphen."
    e3 = "not a valid snap name. Snap names cannot end with a hyphen."
    e4 = "not a valid snap name. Snap names cannot have two hyphens in a row."
    e5 = "too long (maximum length is 40)"

    scenarios = [
        # snapcraft's existing unit tests
        ("existing test #1", dict(name="package@awesome", err=e1)),
        ("existing test #2", dict(name="something.another", err=e1)),
        ("existing test #3", dict(name="_hideme", err=e1)),
        ("existing test #4", dict(name="-no", err=e2)),
        ("existing test #5", dict(name="a:a", err=e1)),
        ("existing test #6", dict(name="123", err=e1)),
        # this one manages to fail every validation test except type
        ("order check", dict(name="-----------------------------------------", err=e5)),
        # from snapd's unit tests (except those covered by above)
        ("name cannot be empty", dict(name="", err=e1)),
        (
            "name cannot be too long",
            dict(name="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa", err=e5),
        ),
        ("dashes alone are not a name", dict(name="-", err=e1)),
        ("dashes alone are not a name, take 2", dict(name="--", err=e1)),
        ("double dashes in a name are not ok", dict(name="a--a", err=e4)),
        ("name should not end with a dash", dict(name="a-", err=e3)),
        ("name cannot have any spaces in it, #1", dict(name="a ", err=e1)),
        ("name cannot have any spaces in it, #2", dict(name=" a", err=e1)),
        ("name cannot have any spaces in it, #3", dict(name="a a", err=e1)),
        ("a number alone is not a name", dict(name="0", err=e1)),
        ("just numbers and dashes", dict(name="1-2-3", err=e1)),
        ("plain ASCII #1", dict(name="реасе", err=e1)),
        ("plain ASCII #2", dict(name="日本語", err=e1)),
        ("plain ASCII #3", dict(name="한글", err=e1)),
        ("plain ASCII #4", dict(name="ру́сский язы́к", err=e1)),
        # from review-tools', except as covered
        (
            "stress the regexper",
            dict(name="u-9490371368748654323415773467328453675-", err=e3),
        ),
        ("review-tools bad", dict(name="foo?bar", err=e1)),
        ("review-tools bad1", dict(name="foo/bar", err=e1)),
        ("review-tools bad6", dict(name="foo-Bar", err=e1)),
    ]

    def test(self, data, name, err):
        data["name"] = name

        with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
            Validator(data).validate()

        assert str(error.value).endswith(
            f"The 'name' property does not match the required schema: {name!r} is {err}"
        )


@pytest.mark.parametrize("snap_type", ["app", "base", "gadget", "kernel", "snapd"])
def test_valid_types(data, snap_type):
    data["type"] = snap_type
    if snap_type in ("base", "kernel", "snapd"):
        data.pop("base")

    Validator(data).validate()


_BASE_TYPE_MSG = (
    "must be one of base: <base> and type: <app|gadget>, "
    "base: bare (with a build-base), "
    "or type: <base|kernel|snapd> (without a base)"
)
_TYPE_ENUM_TMPL = (
    "The 'type' property does not match the required schema: '{}' is not one of "
    "['app', 'base', 'gadget', 'kernel', 'snapd']"
)


@pytest.mark.parametrize("snap_type", ["apps", "framework", "platform", "oem"])
def test_invalid_types(data, snap_type):
    data["type"] = snap_type

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError):
        Validator(data).validate()


def test_type_base_and_no_base(data):
    data.pop("base")
    data["type"] = "base"

    Validator(data).validate()


def test_type_base_and_base(data):
    data["type"] = "base"

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    assert _BASE_TYPE_MSG in str(error.value)


def test_build_base_and_base(data):
    data["build-base"] = "fake-base"

    Validator(data).validate()


def test_build_base_and_type_base(data):
    data.pop("base")
    data["type"] = "base"
    data["build-base"] = "fake-base"

    Validator(data).validate()


def test_build_base_and_base_bare(data):
    data["base"] = "bare"
    data["build-base"] = "fake-base"

    Validator(data).validate()


@pytest.mark.parametrize(
    "name",
    [
        "1",
        "a",
        "aa",
        "aaa",
        "aaaa",
        "Aa",
        "aA",
        "1a",
        "a1",
        "1-a",
        "a-1",
        "a-a",
        "aa-a",
        "a-aa",
        "a-b-c",
        "0a-a",
        "a-0a",
    ],
)
def test_valid_app_names(data, name):
    data["apps"] = {name: {"command": "foo"}}

    Validator(data).validate()


@pytest.mark.parametrize(
    "name",
    [
        "qwe#rty",
        "qwe_rty",
        "queue rty",
        "queue  rty",
        "",
        "-",
        "--",
        "a--a",
        "a-",
        "a ",
        " a",
        "a a",
        "日本語",
        "한글",
        "ру́сский язы́к",
        "ໄຂ່​ອີ​ສ​ເຕີ້",
        ":a",
        "a:",
        "a:a",
        "_a",
        "a_",
        "a_a",
    ],
)
def test_invalid_app_names(data, name):
    data["apps"] = {name: {"command": "1"}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        f"The 'apps' property does not match the required schema: {name!r} is "
        "not a valid app name. App names consist of upper- and lower-case "
        "alphanumeric characters and hyphens. They cannot start or end "
        "with a hyphen."
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize(
    "condition",
    [
        "always",
        "on-success",
        "on-failure",
        "on-abnormal",
        "on-abort",
        "on-watchdog",
        "never",
    ],
)
def test_valid_restart_conditions(data, condition):
    data["apps"] = {"service1": {"command": "binary1", "daemon": "simple"}}
    data["apps"]["service1"]["restart-condition"] = condition

    Validator(data).validate()


_REFRESH_MODES = ["endure", "restart"]


@pytest.mark.parametrize("mode", _REFRESH_MODES)
def test_valid_refresh_modes(data, mode):
    data["apps"] = {
        "service1": {"command": "binary1", "daemon": "simple", "refresh-mode": mode}
    }
    Validator(data).validate()


@pytest.mark.parametrize("mode", _REFRESH_MODES)
def test_refresh_mode_daemon_missing_errors(data, mode):
    data["apps"] = {"service1": {"command": "binary1", "refresh-mode": mode}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError):
        Validator(data).validate()


_STOP_MODES = [
    "sigterm",
    "sigterm-all",
    "sighup",
    "sighup-all",
    "sigusr1",
    "sigusr1-all",
    "sigusr2",
    "sigusr2-all",
]


@pytest.mark.parametrize("mode", _STOP_MODES)
def test_valid_modes(data, mode):
    data["apps"] = {
        "service1": {"command": "binary1", "daemon": "simple", "stop-mode": mode}
    }
    Validator(data).validate()


@pytest.mark.parametrize("mode", _STOP_MODES)
def test_daemon_missing_errors(data, mode):
    data["apps"] = {"service1": {"command": "binary1", "stop-mode": mode}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError):
        Validator(data).validate()


@pytest.mark.parametrize(
    "name",
    [
        "qwe#rty",
        "qwe_rty",
        "queue rty",
        "queue  rty",
        "Hi",
        "",
        "-",
        "--",
        "a--a",
        "a-",
        "a ",
        " a",
        "a a",
        "日本語",
        "한글",
        "ру́сский язы́к",
        "ໄຂ່​ອີ​ສ​ເຕີ້",
        ":a",
        "a:",
        "a:a",
        "_a",
        "a_",
        "a_a",
    ],
)
def test_invalid_hook_names(data, name):
    data["hooks"] = {name: {"plugs": ["network"]}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        f"The 'hooks' property does not match the required schema: {name!r} is "
        "not a valid hook name. Hook names consist of lower-case "
        "alphanumeric characters and hyphens. They cannot start or end "
        "with a hyphen."
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize(
    "name", ["plugins", "qwe#rty", "qwe_rty", "queue rty", "queue  rty", "part/sub"]
)
def test_invalid_part_names(data, name):
    data["parts"] = {name: {"plugin": "nil"}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        f"The 'parts' property does not match the required schema: {name!r} is "
        "not a valid part name. Part names consist of lower-case "
        "alphanumeric characters, hyphens and plus signs. "
        "As a special case, 'plugins' is also not a valid part name."
    )
    assert str(error.value).endswith(expected_message)


class TestInvalidArchitectures:

    scenarios = [
        (
            "single string",
            {"architectures": "amd64", "message": "'amd64' is not of type 'array'"},
        ),
        (
            "unknown object properties",
            {
                "architectures": [{"builds-on": ["amd64"], "runs-on": ["amd64"]}],
                "message": "'build-on' is a required property and additional "
                "properties are not allowed",
            },
        ),
        (
            "omit build-on",
            {
                "architectures": [{"run-on": ["amd64"]}],
                "message": "'build-on' is a required property",
            },
        ),
        (
            "build on all and others",
            {
                "architectures": [{"build-on": ["amd64", "all"]}],
                "message": "'all' can only be used within 'build-on' by itself, "
                "not with other architectures",
            },
        ),
        (
            "run on all and others",
            {
                "architectures": [{"build-on": ["amd64"], "run-on": ["amd64", "all"]}],
                "message": "'all' can only be used within 'run-on' by itself, "
                "not with other architectures",
            },
        ),
        (
            "run on all and more objects",
            {
                "architectures": [
                    {"build-on": ["amd64"], "run-on": ["all"]},
                    {"build-on": ["i396"], "run-on": ["i386"]},
                ],
                "message": "one of the items has 'all' in 'run-on', but there are "
                "2 items: upon release they will conflict. 'all' "
                "should only be used if there is a single item",
            },
        ),
        (
            "build on all and more objects",
            {
                "architectures": [{"build-on": ["all"]}, {"build-on": ["i396"]}],
                "message": "one of the items has 'all' in 'build-on', but there "
                "are 2 items: snapcraft doesn't know which one to use. "
                "'all' should only be used if there is a single item",
            },
        ),
        (
            "multiple builds run on same arch",
            {
                "architectures": [
                    {"build-on": ["amd64"], "run-on": ["amd64"]},
                    {"build-on": ["i396"], "run-on": ["amd64", "i386"]},
                ],
                "message": "multiple items will build snaps that claim to run on "
                "'amd64'",
            },
        ),
        (
            "multiple builds run on same arch with implicit run-on",
            {
                "architectures": [
                    {"build-on": ["amd64"]},
                    {"build-on": ["i396"], "run-on": ["amd64", "i386"]},
                ],
                "message": "multiple items will build snaps that claim to run on "
                "'amd64'",
            },
        ),
        (
            "mixing forms",
            {
                "architectures": [
                    "amd64",
                    {"build-on": ["i386"], "run-on": ["amd64", "i386"]},
                ],
                "message": "every item must either be a string or an object",
            },
        ),
        (
            "build on all run on specific",
            {
                "architectures": [
                    {"build-on": ["all"], "run-on": ["amd64"]},
                    {"build-on": ["all"], "run-on": ["i386"]},
                ],
                "message": "one of the items has 'all' in 'build-on', but there "
                "are 2 items: snapcraft doesn't know which one to use. "
                "'all' should only be used if there is a single item",
            },
        ),
        (
            "build on overlap",
            {
                "architectures": [
                    {"build-on": ["amd64", "i386"], "run-on": ["i386"]},
                    {"build-on": ["amd64"], "run-on": ["amd64"]},
                ],
                "message": "'amd64' is present in the 'build-on' of multiple "
                "items, which means snapcraft doesn't know which "
                "'run-on' to use when building on that architecture",
            },
        ),
    ]

    def test(self, data, architectures, message):
        data["architectures"] = architectures

        with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
            Validator(data).validate()

        assert message in str(error.value)


class IconTest(ProjectBaseTest):
    def test_invalid_yaml_invalid_icon_extension(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            icon: icon.foo
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: nil
            """
            )
        )

        self.assertThat(
            raised.message, Equals("icon 'icon.foo' must be either a .png or a .svg")
        )


@pytest.mark.parametrize("title", ["a title", "A Title", "T" * 40, "💩" * 40, "’" * 40])
def test_valid_title(data, title):
    data["title"] = title

    Validator(data).validate()


_EXPECTED_ERROR_TEMPLATE = {
    "number": (
        "Issues while validating snapcraft.yaml: The 'title' property "
        "does not match the required schema: {} is not of type 'string'"
    ),
    "length": (
        "Issues while validating snapcraft.yaml: The 'title' property "
        "does not match the required schema: {!r} is too long "
        "(maximum length is 40)"
    ),
}


@pytest.mark.parametrize(
    "title,error_template",
    [(1, "number"), ("T" * 41, "length"), ("💩" * 41, "length"), ("’" * 41, "length")],
)
def test_invalid_title(data, title, error_template):
    data["title"] = title

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    assert _EXPECTED_ERROR_TEMPLATE[error_template].format(title) in str(error.value)


class OrganizeTest(ProjectBaseTest):
    def test_yaml_organize_value_none(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: nothing
            confinement: strict

            parts:
              part1:
                plugin: nil
                organize:
                  foo:
            """
            )
        )

        self.assertThat(
            str(raised),
            Contains(
                "The 'parts/part1/organize/foo' property does not match the "
                "required schema: None is not of type 'string'"
            ),
        )

    def test_yaml_organize_value_empty(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: nothing
            confinement: strict

            parts:
              part1:
                plugin: nil
                organize:
                  foo: ''
            """
            )
        )

        self.assertThat(
            str(raised),
            Contains(
                "The 'parts/part1/organize/foo' property does not match the "
                "required schema: '' is too short (minimum length is 1)"
            ),
        )


class VersionTest(ProjectBaseTest):
    def test_invalid_yaml_version_too_long(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: 'abcdefghijklmnopqrstuvwxyz1234567' # Max is 32 in the store
            summary: test
            description: test
            confinement: strict
            grade: stable
            parts:
              part1:
                plugin: nil
        """
            )
        )  # noqa: E501

        self.assertThat(
            raised.message,
            Equals(
                "The 'version' property does not match the required "
                "schema: 'abcdefghijklmnopqrstuvwxyz1234567' is too long "
                "(maximum length is 32)"
            ),
        )


@pytest.mark.parametrize(
    "version", ["buttered-popcorn", "1.2.3", "v12.4:1:2~", "HeLlo", "v+"]
)
def test_valid_version(data, version):
    data["version"] = version

    Validator(data).validate()


@pytest.mark.parametrize(
    "version",
    [
        "'*'",
        "''",
        "':v'",
        "'.v'",
        "'+v'",
        "'~v'",
        "'_v'",
        "'-v'",
        "'v:'",
        "'v.'",
        "'v_'",
        "'v-'",
        "'underscores_are_bad'",
    ],
)
def test_invalid_version(data, version):
    data["version"] = version

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'version' property does not match the required "
        f"schema: {version!r} is not a valid snap version string. Snap versions "
        "consist of upper- and lower-case alphanumeric characters, "
        "as well as periods, colons, plus signs, tildes, and "
        "hyphens. They cannot begin with a period, colon, plus "
        "sign, tilde, or hyphen. They cannot end with a period, "
        "colon, or hyphen."
    )
    assert expected_message in str(error.value)


def test_invalid_version_type(data):
    data["version"] = 0.1

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'version' property does not match the required "
        "schema: snap versions need to be strings. They must "
        "also be wrapped in quotes when the value will be "
        "interpreted by the YAML parser as a non-string. "
        "Examples: '1', '1.2', '1.2.3', git (will be replaced "
        "by a git describe based version string)."
    )
    assert expected_message in str(error.value)


def test_invalid_version_length(data):
    data["version"] = "this.is.a.really.too.long.version"

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'version' property does not match the required "
        "schema: 'this.is.a.really.too.long.version' is too long "
        "(maximum length is 32)"
    )
    assert expected_message in str(error.value)


class EnvironmentTest(ProjectBaseTest):
    def test_valid_environment(self):
        snapcraft_yaml = self.assertValidationPasses(
            dedent(
                dedent(
                    """\
            name: project-name
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            environment:
                GLOBAL: "1"
                OTHER: valid-value

            apps:
              app1:
                command: app1
                environment:
                  LOCALE: C
                  PLUGIN_PATH: $SNAP_USER_DATA/plugins

            parts:
              main:
                plugin: nil
                source: .
            """
                )
            )
        )

        self.assertThat(
            snapcraft_yaml["environment"], Equals(dict(GLOBAL="1", OTHER="valid-value"))
        )
        self.assertThat(
            snapcraft_yaml["apps"]["app1"]["environment"],
            Equals(dict(LOCALE="C", PLUGIN_PATH="$SNAP_USER_DATA/plugins")),
        )

    def test_invalid_environment(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: project-name
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            environment:
                INVALID:
                    - 1
                    - 2

            parts:
              main:
                source: .
                plugin: nil
            """
            )
        )

        self.assertRegex(
            raised.message,
            r"The 'environment/INVALID' property does not match the required "
            r"schema: \[1, 2\].*",
        )


@pytest.mark.parametrize("compression", ["lzo", "xz"])
def test_valid_compression(data, compression):
    data["compression"] = compression

    Validator(data).validate()


@pytest.mark.parametrize("compression", ["lzma", "gz", "rar"])
def test_invalid_compression(data, compression):
    data["compression"] = compression

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'compression' property does not match the required "
        f"schema: {compression!r} is not one of ['lzo', 'xz']"
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize("confinement", ["strict", "devmode", "classic"])
def test_valid_confinement(data, confinement):
    data["confinement"] = confinement

    Validator(data).validate()


@pytest.mark.parametrize("confinement", ["foo", "strict-", "_devmode"])
def test_invalid_confinement(data, confinement):
    data["confinement"] = confinement

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'confinement' property does not match the required "
        f"schema: {confinement!r} is not one of ['classic', 'devmode', "
        "'strict']"
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize("desc", ["test", "multi\nline\n"])
def test_valid_description(data, desc):
    data["description"] = desc
    Validator(data).validate()


@pytest.mark.parametrize("desc", [""])
def test_invalid_description(data, desc):
    data["description"] = desc

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'description' property does not match the required "
        f"schema: {desc!r} is not a valid description string"
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize("grade", ["stable", "devel"])
def test_valid_grade(data, grade):
    data["grade"] = grade

    Validator(data).validate()


@pytest.mark.parametrize("grade", ["foo", "strict-", "_devmode"])
def test_invalid_grade(data, grade):
    data["grade"] = grade

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'grade' property does not match the required "
        f"schema: {grade!r} is not one of ['stable', 'devel']"
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize("epoch", [0, "0", "1*", 1, "1", "400*", "1234"])
def test_valid_epoch(data, epoch):
    data["epoch"] = epoch

    Validator(data).validate()


@pytest.mark.parametrize(
    "epoch",
    [
        "0*",
        "_",
        "1-",
        "1+",
        "-1",
        "-1*",
        "a",
        "1a",
        "1**",
        '"01"',
        "1.2",
        '"1.2"',
        "[1]",
    ],
)
def test_invalid_epoch(data, epoch):
    data["epoch"] = epoch

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        f"{epoch!r} is not a 'epoch' (epochs are positive integers "
        "followed by an optional asterisk)"
    )
    assert expected_message in str(error.value)


def test_valid_license(data):
    data["license"] = "MIT-0"

    Validator(data).validate()


def test_invalid_license(data):
    data["license"] = 1234

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = (
        "The 'license' property does not match the required schema: "
        "1234 is not of type 'string'"
    )
    assert expected_message in str(error.value)


@pytest.mark.parametrize("adapter", ["none", "legacy", "full"])
def test_valid_adapter(data, adapter):
    data["apps"] = {"foo": {"command": "foo", "adapter": adapter}}

    Validator(data).validate()


@pytest.mark.parametrize("adapter", ["NONE", "F", "Full"])
def test_invalid_adapter(data, adapter):
    data["apps"] = {"foo": {"command": "foo", "adapter": adapter}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = "The 'apps/foo/adapter' property does not match"
    assert expected_message in str(error.value)


@pytest.mark.parametrize(
    "build_environment",
    ["a string", ["a string"], [{"k1": "v1", "k2": "v2"}], [{"k1": 5}]],
)
def test_invalid_part_build_environment_key_type(data, build_environment):
    data["parts"]["part1"]["build-environment"] = build_environment

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError):
        Validator(data).validate()


@pytest.mark.parametrize(
    "command_chain", ["a string", [1], ["test chain"], ["test'chain"]]
)
def test_invalid_command_chain(data, command_chain):
    data["apps"] = {"foo": {"command": "foo", "command-chain": command_chain}}

    with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
        Validator(data).validate()

    expected_message = "The 'apps/foo/command-chain"
    assert expected_message in str(error.value)


class SystemUsernamesTests(ProjectBaseTest):
    def test_yaml_valid_system_usernames_long(self):
        self.assertValidationPasses(
            dedent(
                """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: nothing
                license: MIT
                parts:
                  part1:
                    plugin: nil
                system-usernames:
                  snap_daemon:
                    scope: shared
                """
            )
        )

    def test_yaml_valid_system_usernames_short(self):
        self.assertValidationPasses(
            dedent(
                """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: nothing
                license: MIT
                parts:
                  part1:
                    plugin: nil
                system-usernames:
                  snap_daemon: shared
                """
            )
        )

    def test_invalid_yaml_invalid_username(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: nothing
            license: MIT
            parts:
              part1:
                plugin: nil
            system-usernames:
              snap_user: shared
            """
            )
        )

        self.assertThat(
            raised.message,
            Equals(
                "The 'system-usernames' property does not match the required schema: 'snap_user' is not a valid system-username."
            ),
        )

    def test_invalid_yaml_invalid_short_scope(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: nothing
            license: MIT
            parts:
              part1:
                plugin: nil
            system-usernames:
              snap_daemon: invalid-scope
            """
            )
        )

        self.assertThat(
            raised.message,
            Equals(
                "The 'system-usernames/snap_daemon' property does not match the required schema: 'invalid-scope' is not valid under any of the given schemas"
            ),
        )

    def test_invalid_yaml_invalid_long_scope(self):
        raised = self.assertValidationRaises(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: nothing
            license: MIT
            parts:
              part1:
                plugin: nil
            system-usernames:
              snap_daemon:
                scope: invalid-scope
            """
            )
        )

        self.assertThat(
            raised.message,
            Equals(
                "The 'system-usernames/snap_daemon' property does not match the required schema: OrderedDict([('scope', 'invalid-scope')]) is not valid under any of the given schemas"
            ),
        )


class PackageManagement(ProjectBaseTest):
    def test_yaml_valid_apt_repositories(self):
        self.assertValidationPasses(
            dedent(
                """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: nothing
                license: MIT
                parts:
                  part1:
                    plugin: nil
                package-repositories:
                  - type: apt
                    ppa: user/test-ppa
                  - type: apt
                    architectures: [amd64, i386]
                    components: [main, multiverse]
                    formats: [deb, deb-src]
                    key-id: test-key-id
                    key-server: keyserver.ubuntu.com
                    url: http://archive.ubuntu.com/ubuntu
                    suites: [test, test-updates, test-security]
                  - type: apt
                    key-id: test-key-id
                    url: http://archive.ubuntu.com/ubuntu
                    suites: [bionic, bionic-updates]
                  - type: apt
                    key-id: test-key-id
                    path: foo
                    url: http://archive.ubuntu.com/ubuntu
                """
            )
        )


class TestInvalidAptConfigurations:
    scenarios = [
        (
            "ppa extra invalid field",
            dict(
                packages=[{"type": "apt", "ppa": "test/ppa", "invalid": "invalid"}],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb extra invalid field",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                        "invalid": "invalid",
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb missing field: key-id",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb missing field: url",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "suites": ["test", "test-updates", "test-security"],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb invalid deb type",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "format": ["invalid"],
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb invalid key-id",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "\\*\\*",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb empty architectures",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                        "architectures": [],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "deb empty suites",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": [],
                    }
                ],
                message_contains="The 'package-repositories[0]' property does not match the required schema:",
            ),
        ),
        (
            "ppa duplicate",
            dict(
                packages=[
                    {"type": "apt", "ppa": "dupe/check"},
                    {"type": "apt", "ppa": "dupe/check"},
                ],
                message_contains="has non-unique elements",
            ),
        ),
        (
            "deb duplicate",
            dict(
                packages=[
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                    },
                    {
                        "type": "apt",
                        "components": ["main"],
                        "key-id": "test-key-id",
                        "url": "http://archive.ubuntu.com/ubuntu",
                        "suites": ["test", "test-updates", "test-security"],
                    },
                ],
                message_contains="has non-unique elements",
            ),
        ),
    ]

    def test_invalid(self, data, packages, message_contains):
        data["package-repositories"] = packages

        with pytest.raises(snapcraft.yaml_utils.errors.YamlValidationError) as error:
            Validator(data).validate()

        assert message_contains in str(error.value)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
from typing import Optional
from unittest import mock

import fixtures

from snapcraft.internal import os_release


class FakeOsRelease(fixtures.Fixture):
    def __init__(
        self,
        id: str = "ubuntu",
        version_id: Optional[str] = "16.04",
        version_codename: Optional[str] = None,
        name: Optional[str] = "Ubuntu",
    ) -> None:
        self._id = id
        self._version_id = version_id
        self._version_codename = version_codename
        self._name = name

    def _setUp(self):
        super()._setUp()

        with open("os-release", "w") as release_file:
            if self._name is not None:
                print(f'NAME="{self._name}"', file=release_file)

            print(
                dedent(
                    f"""\
                VERSION="16.04.3 LTS (Xenial Xerus)"
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 16.04.3 LTS"
                HOME_URL="http://www.ubuntu.com/"
                SUPPORT_URL="http://help.ubuntu.com/"
                BUG_REPORT_URL="http://bugs.launchpad.net/ubuntu/"
                UBUNTU_CODENAME=xenial"""
                ),
                file=release_file,
            )

            if self._id is not None:
                print("ID={}".format(self._id), file=release_file)
            if self._version_id is not None:
                print('VERSION_ID="{}"'.format(self._version_id), file=release_file)
            if self._version_codename is not None:
                print(
                    "VERSION_CODENAME={}".format(self._version_codename),
                    file=release_file,
                )

        release = os_release.OsRelease(os_release_file="os-release")

        def _create_os_release(*args, **kwargs):
            return release

        patcher = mock.patch(
            "snapcraft.internal.os_release.OsRelease", wraps=_create_os_release
        )
        patcher.start()
        self.addCleanup(patcher.stop)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import logging
import os
import os.path
import pathlib
import re
from typing import List, Set

import jsonschema

from snapcraft import formatting_utils, plugins, project
from snapcraft.internal import deprecations, repo, states, steps
from snapcraft.internal.meta.package_repository import PackageRepository
from snapcraft.internal.meta.snap import Snap
from snapcraft.internal.pluginhandler._part_environment import (
    get_snapcraft_global_environment,
)
from snapcraft.internal.repo import apt_key_manager, apt_sources_manager
from snapcraft.project._schema import Validator

from . import errors, grammar_processing, replace_attr
from ._env import environment_to_replacements, runtime_env
from ._extensions import apply_extensions
from ._parts_config import PartsConfig

logger = logging.getLogger(__name__)


@jsonschema.FormatChecker.cls_checks("icon-path")
def _validate_icon(icon):
    allowed_extensions = [".png", ".svg"]
    extension = os.path.splitext(icon.lower())[1]
    if extension not in allowed_extensions:
        raise jsonschema.exceptions.ValidationError(
            f"icon {icon!r} must be either a .png or a .svg"
        )

    return True


@jsonschema.FormatChecker.cls_checks("epoch", raises=errors.InvalidEpochError)
def _validate_epoch(instance):
    str_instance = str(instance)
    pattern = re.compile("^(?:0|[1-9][0-9]*[*]?)$")
    if not pattern.match(str_instance):
        raise errors.InvalidEpochError()

    return True


@jsonschema.FormatChecker.cls_checks("architectures")
def _validate_architectures(instance):
    standalone_build_ons = collections.Counter()
    build_ons = collections.Counter()
    run_ons = collections.Counter()

    saw_strings = False
    saw_dicts = False

    for item in instance:
        # This could either be a dict or a string. In the latter case, the
        # schema will take care of it. We just need to further validate the
        # dict.
        if isinstance(item, str):
            saw_strings = True
        elif isinstance(item, dict):
            saw_dicts = True
            build_on = _get_architectures_set(item, "build-on")
            build_ons.update(build_on)

            # Add to the list of run-ons. However, if no run-on is specified,
            # we know it's implicitly the value of build-on, so use that
            # for validation instead.
            run_on = _get_architectures_set(item, "run-on")
            if run_on:
                run_ons.update(run_on)
            else:
                standalone_build_ons.update(build_on)

    # Architectures can either be a list of strings, or a list of objects.
    # Mixing the two forms is unsupported.
    if saw_strings and saw_dicts:
        raise jsonschema.exceptions.ValidationError(
            "every item must either be a string or an object",
            path=["architectures"],
            instance=instance,
        )

    # At this point, individual build-ons and run-ons have been validated,
    # we just need to validate them across each other.

    # First of all, if we have a `run-on: [all]` (or a standalone
    # `build-on: [all]`) then we should only have one item in the instance,
    # otherwise we know we'll have multiple snaps claiming they run on the same
    # architectures (i.e. all and something else).
    number_of_snaps = len(instance)
    if "all" in run_ons and number_of_snaps > 1:
        raise jsonschema.exceptions.ValidationError(
            "one of the items has 'all' in 'run-on', but there are {} "
            "items: upon release they will conflict. 'all' should only be "
            "used if there is a single item".format(number_of_snaps),
            path=["architectures"],
            instance=instance,
        )
    if "all" in build_ons and number_of_snaps > 1:
        raise jsonschema.exceptions.ValidationError(
            "one of the items has 'all' in 'build-on', but there are {} "
            "items: snapcraft doesn't know which one to use. 'all' should "
            "only be used if there is a single item".format(number_of_snaps),
            path=["architectures"],
            instance=instance,
        )

    # We want to ensure that multiple `run-on`s (or standalone `build-on`s)
    # don't include the same arch, or they'll clash with each other when
    # releasing.
    all_run_ons = run_ons + standalone_build_ons
    duplicates = {arch for (arch, count) in all_run_ons.items() if count > 1}
    if duplicates:
        raise jsonschema.exceptions.ValidationError(
            "multiple items will build snaps that claim to run on {}".format(
                formatting_utils.humanize_list(duplicates, "and")
            ),
            path=["architectures"],
            instance=instance,
        )

    # Finally, ensure that multiple `build-on`s don't include the same arch
    # or Snapcraft has no way of knowing which one to use.
    duplicates = {arch for (arch, count) in build_ons.items() if count > 1}
    if duplicates:
        raise jsonschema.exceptions.ValidationError(
            "{} {} present in the 'build-on' of multiple items, which means "
            "snapcraft doesn't know which 'run-on' to use when building on "
            "{} {}".format(
                formatting_utils.humanize_list(duplicates, "and"),
                formatting_utils.pluralize(duplicates, "is", "are"),
                formatting_utils.pluralize(duplicates, "that", "those"),
                formatting_utils.pluralize(duplicates, "architecture", "architectures"),
            ),
            path=["architectures"],
            instance=instance,
        )

    return True


def _get_architectures_set(item, name):
    value = item.get(name, set())
    if isinstance(value, str):
        value_set = {value}
    else:
        value_set = set(value)

    _validate_architectures_set(value_set, name)

    return value_set


def _validate_architectures_set(architectures_set, name):
    if "all" in architectures_set and len(architectures_set) > 1:
        raise jsonschema.exceptions.ValidationError(
            "'all' can only be used within {!r} by itself, "
            "not with other architectures".format(name),
            path=["architectures"],
            instance=architectures_set,
        )


class Config:
    @property
    def part_names(self):
        return self.parts.part_names

    @property
    def all_parts(self):
        return self.parts.all_parts

    def __init__(self, project: project.Project) -> None:
        self.build_snaps: Set[str] = set()
        self.project = project

        # raw_snapcraft_yaml is read only, create a new copy
        snapcraft_yaml = apply_extensions(project.info.get_raw_snapcraft())

        self.validator = Validator(snapcraft_yaml)
        self.validator.validate()

        snapcraft_yaml = self._expand_filesets(snapcraft_yaml)

        self.data = self._expand_env(snapcraft_yaml)

        self.data["architectures"] = _process_architectures(
            self.data.get("architectures"), project.deb_arch
        )

        self._ensure_no_duplicate_app_aliases()

        self._global_grammar_processor = grammar_processing.GlobalGrammarProcessor(
            properties=self.data, project=project
        )

        # XXX: Resetting snap_meta due to above mangling of data.
        # Convergence to operating on snap_meta will remove this requirement...
        project._snap_meta = Snap.from_dict(self.data)

        self.parts = PartsConfig(
            parts=self.data, project=project, validator=self.validator
        )

    def _ensure_no_duplicate_app_aliases(self):
        # Prevent multiple apps within a snap from having duplicate alias names
        aliases = []
        for app_name, app in self.data.get("apps", {}).items():
            aliases.extend(app.get("aliases", []))

        # The aliases property is actually deprecated:
        if aliases:
            deprecations.handle_deprecation_notice("dn5")
        seen = set()
        duplicates = set()
        for alias in aliases:
            if alias in seen:
                duplicates.add(alias)
            else:
                seen.add(alias)
        if duplicates:
            raise errors.DuplicateAliasError(aliases=duplicates)

    def _get_required_package_repositories(self) -> List[PackageRepository]:
        package_repos = self.project._snap_meta.package_repositories.copy()

        v1_plugins = [
            part.plugin
            for part in self.all_parts
            if isinstance(part.plugin, plugins.v1.PluginV1)
        ]
        for plugin in v1_plugins:
            package_repos.extend(plugin.get_required_package_repositories())

        return package_repos

    def _verify_all_key_assets_installed(
        self, *, key_assets: pathlib.Path, key_manager: apt_key_manager.AptKeyManager,
    ) -> None:
        """Verify all configured key assets are utilized, error if not."""
        for key_asset in key_assets.glob("*"):
            key = key_asset.read_text()
            for key_id in key_manager.get_key_fingerprints(key=key):
                if not key_manager.is_key_installed(key_id=key_id):
                    raise errors.SnapcraftProjectUnusedKeyAssetError(key_path=key_asset)

    def install_package_repositories(self) -> None:
        package_repos = self._get_required_package_repositories()
        if not package_repos:
            return

        # Install pre-requisite packages for apt-key, if not installed.
        repo.Repo.install_build_packages(package_names=["gnupg", "dirmngr"])

        key_assets = self.project._get_keys_path()
        key_manager = apt_key_manager.AptKeyManager(key_assets=key_assets)
        sources_manager = apt_sources_manager.AptSourcesManager()

        refresh_required = False
        for package_repo in self._get_required_package_repositories():
            refresh_required |= key_manager.install_package_repository_key(
                package_repo=package_repo
            )
            refresh_required |= sources_manager.install_package_repository_sources(
                package_repo=package_repo
            )

        self._verify_all_key_assets_installed(
            key_assets=key_assets, key_manager=key_manager
        )

        if refresh_required:
            repo.Repo.refresh_build_packages()

    def get_build_packages(self) -> Set[str]:
        # Install/update configured package repositories.
        self.install_package_repositories()

        build_packages = self._global_grammar_processor.get_build_packages()
        build_packages |= set(self.project.additional_build_packages)

        if self.project._snap_meta.version == "git":
            build_packages.add("git")

        for part in self.all_parts:
            build_packages |= part._grammar_processor.get_build_packages()

            # TODO: this should not pass in command but the required package,
            #       where the required package is to be determined by the
            #       source handler.
            if part.source_handler and part.source_handler.command:
                # TODO get_packages_for_source_type should not be a thing.
                build_packages |= repo.Repo.get_packages_for_source_type(
                    part.source_handler.command
                )

            if not isinstance(part.plugin, plugins.v1.PluginV1):
                build_packages |= part.plugin.get_build_packages()

        return build_packages

    def get_build_snaps(self) -> Set[str]:
        build_snaps = set()

        # Add the base.
        if self.project._snap_meta.base is not None:
            build_snaps.add(self.project._snap_meta.base)

        for part in self.all_parts:
            build_snaps |= part._grammar_processor.get_build_snaps()
            if not isinstance(part.plugin, plugins.v1.PluginV1):
                build_snaps |= part.plugin.get_build_snaps()

        return build_snaps

    def get_project_state(self, step: steps.Step):
        """Returns a dict of states for the given step of each part."""

        state = {}
        for part in self.parts.all_parts:
            state[part.name] = states.get_state(part.part_state_dir, step)

        return state

    def snap_env(self):
        prime_dir = self.project.prime_dir
        env = []

        env += runtime_env(prime_dir, self.project.arch_triplet)
        dependency_paths = set()
        for part in self.parts.all_parts:
            env += part.env(prime_dir)
            dependency_paths |= part.get_primed_dependency_paths()

        # Dependency paths are only valid if they actually exist. Sorting them
        # here as well so the LD_LIBRARY_PATH is consistent between runs.
        dependency_paths = sorted(
            {path for path in dependency_paths if os.path.isdir(path)}
        )

        if dependency_paths:
            # Add more specific LD_LIBRARY_PATH from the dependencies.
            env.append(
                'LD_LIBRARY_PATH="'
                + ":".join(dependency_paths)
                + '${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"'
            )

        return env

    def project_env(self):
        return [
            '{}="{}"'.format(variable, value)
            for variable, value in get_snapcraft_global_environment(
                self.project
            ).items()
        ]

    def _expand_env(self, snapcraft_yaml):
        environment_keys = ["name", "version"]
        for key in snapcraft_yaml:
            if any((key == env_key for env_key in environment_keys)):
                continue

            replacements = environment_to_replacements(
                get_snapcraft_global_environment(self.project)
            )

            snapcraft_yaml[key] = replace_attr(snapcraft_yaml[key], replacements)
        return snapcraft_yaml

    def _expand_filesets(self, snapcraft_yaml):
        parts = snapcraft_yaml.get("parts", {})

        for part_name in parts:
            for step in ("stage", "prime"):
                step_fileset = _expand_filesets_for(step, parts[part_name])
                parts[part_name][step] = step_fileset

        return snapcraft_yaml


def _expand_filesets_for(step, properties):
    filesets = properties.get("filesets", {})
    fileset_for_step = properties.get(step, {})
    new_step_set = []

    for item in fileset_for_step:
        if item.startswith("$"):
            try:
                new_step_set.extend(filesets[item[1:]])
            except KeyError:
                raise errors.SnapcraftLogicError(
                    "'{}' referred to in the '{}' fileset but it is not "
                    "in filesets".format(item, step)
                )
        else:
            new_step_set.append(item)

    return new_step_set


class _Architecture:
    def __init__(self, *, build_on, run_on=None):
        if isinstance(build_on, str):
            self.build_on = [build_on]
        else:
            self.build_on = build_on

        # If there is no run_on, it defaults to the value of build_on
        if not run_on:
            self.run_on = self.build_on
        elif isinstance(run_on, str):
            self.run_on = [run_on]
        else:
            self.run_on = run_on


def _create_architecture_list(architectures, current_arch):
    if not architectures:
        return [_Architecture(build_on=[current_arch])]

    build_architectures: List[str] = []
    architecture_list: List[_Architecture] = []
    for item in architectures:
        if isinstance(item, str):
            build_architectures.append(item)
        if isinstance(item, dict):
            architecture_list.append(
                _Architecture(build_on=item.get("build-on"), run_on=item.get("run-on"))
            )

    if build_architectures:
        architecture_list.append(_Architecture(build_on=build_architectures))

    return architecture_list


def _process_architectures(architectures, current_arch):
    architecture_list = _create_architecture_list(architectures, current_arch)

    for architecture in architecture_list:
        if current_arch in architecture.build_on or "all" in architecture.build_on:
            return architecture.run_on

    return [current_arch]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib

import pytest

from snapcraft.project import errors, get_snapcraft_yaml


@pytest.fixture(
    params=[
        "snapcraft.yaml",
        "snap/snapcraft.yaml",
        ".snapcraft.yaml",
        "build-aux/snap/snapcraft.yaml",
    ]
)
def stub_snapcraft_yaml_file(tmp_work_path, request):
    """Return stub snapcraft.yaml file variants to find."""
    snapcraft_yaml = request.param
    snapcraft_yaml_path = tmp_work_path / snapcraft_yaml
    snapcraft_yaml_path.parent.mkdir(parents=True, exist_ok=True)
    snapcraft_yaml_path.touch()
    return snapcraft_yaml


def test_get_snapcraft_yaml(stub_snapcraft_yaml_file):
    assert get_snapcraft_yaml() == stub_snapcraft_yaml_file


def test_config_raises_on_missing_snapcraft_yaml(tmp_work_path):
    """Test that an error is raised if snap/snapcraft.yaml is missing"""
    with pytest.raises(errors.MissingSnapcraftYamlError):
        get_snapcraft_yaml()


@pytest.fixture(
    params=["snap/snapcraft.yaml", ".snapcraft.yaml", "build-aux/snap/snapcraft.yaml"]
)
def duplicate_stub_snapcraft_yaml_file(stub_snapcraft_yaml_file, request):
    """Return duplicate stub snapcraft.yaml file variants to find."""
    snapcraft_yaml = request.param

    if snapcraft_yaml == stub_snapcraft_yaml_file:
        snapcraft_yaml = "snapcraft.yaml"

    snapcraft_yaml_path = pathlib.Path(snapcraft_yaml)
    snapcraft_yaml_path.parent.mkdir(parents=True, exist_ok=True)
    snapcraft_yaml_path.touch()
    return snapcraft_yaml


def test_duplicates(duplicate_stub_snapcraft_yaml_file):
    with pytest.raises(errors.DuplicateSnapcraftYamlError):
        get_snapcraft_yaml()
